import threading
from concurrent.futures import ThreadPoolExecutor

import structlog
from eth_typing import Address

import beamer.metrics
from beamer.chain import POLL_PERIOD_MAINNET, POLL_PERIOD_TESTNET, EventMonitor, EventProcessor
from beamer.config import Config
from beamer.contracts import make_contracts
from beamer.state_machine import Context
from beamer.tracker import Tracker
from beamer.typing import ChainId
from beamer.util import TokenMatchChecker, make_web3

log = structlog.get_logger(__name__)


class Agent:
    def __init__(self, config: Config):
        self._config = config
        self._stopped = threading.Event()
        self._stopped.set()
        # Just add one worker, as this effectively serializes the work
        # This is necessary as we use the account for all resolutions and
        # would otherwise run into nonce problems
        self._task_pool = ThreadPoolExecutor(max_workers=1)

        w3_l1 = make_web3(config.l1_rpc_url, config.account)
        w3_l2a = make_web3(config.l2a_rpc_url, config.account)
        w3_l2b = make_web3(config.l2b_rpc_url, config.account)

        l2a_contracts_info = config.deployment_info[ChainId(w3_l2a.eth.chain_id)]
        l2b_contracts_info = config.deployment_info[ChainId(w3_l2b.eth.chain_id)]

        l2a_contracts = make_contracts(w3_l2a, l2a_contracts_info)
        l2b_contracts = make_contracts(w3_l2b, l2b_contracts_info)

        request_manager = l2a_contracts["RequestManager"]
        resolution_registry = l2a_contracts["ResolutionRegistry"]
        fill_manager = l2b_contracts["FillManager"]

        if not fill_manager.functions.allowedLPs(config.account.address).call():
            raise RuntimeError("Agent address is not whitelisted")

        with open(config.token_match_file, "r") as f:
            match_checker = TokenMatchChecker.from_file(f)

        self.context = Context(
            requests=Tracker(),
            claims=Tracker(),
            request_manager=request_manager,
            fill_manager=fill_manager,
            match_checker=match_checker,
            address=config.account.address,
            latest_blocks={},
            config=config,
            web3_l1=w3_l1,
            task_pool=self._task_pool,
            l1_resolutions={},
        )

        chain_id = w3_l1.eth.chain_id
        poll_period = POLL_PERIOD_MAINNET if chain_id == 1 else POLL_PERIOD_TESTNET

        self._event_processor = EventProcessor(self.context)

        self._event_monitor_l2a = EventMonitor(
            web3=w3_l2a,
            contracts=(request_manager, resolution_registry),
            deployment_block=min(
                l2a_contracts_info["RequestManager"].deployment_block,
                l2a_contracts_info["ResolutionRegistry"].deployment_block,
            ),
            on_new_events=self._event_processor.add_events,
            on_sync_done=self._event_processor.mark_sync_done,
            poll_period=poll_period,
        )

        self._event_monitor_l2b = EventMonitor(
            web3=w3_l2b,
            contracts=(fill_manager,),
            deployment_block=l2b_contracts_info["FillManager"].deployment_block,
            on_new_events=self._event_processor.add_events,
            on_sync_done=self._event_processor.mark_sync_done,
            poll_period=poll_period,
        )

    def start(self) -> None:
        assert self._stopped.is_set()
        beamer.metrics.init(self._config)
        self._event_processor.start()
        self._event_monitor_l2a.start()
        self._event_monitor_l2b.start()
        self._stopped.clear()

    def stop(self) -> None:
        assert not self._stopped.is_set()
        self._event_processor.stop()
        self._event_monitor_l2a.stop()
        self._event_monitor_l2b.stop()
        self._task_pool.shutdown(wait=True, cancel_futures=False)
        self._stopped.set()

    @property
    def running(self) -> bool:
        return not self._stopped.is_set()

    @property
    def address(self) -> Address:
        return self._config.account.address

    def wait(self) -> None:
        self._stopped.wait()



import json
from dataclasses import dataclass
from pathlib import Path

import web3
from web3.contract import Contract

from beamer.typing import Address, BlockNumber, ChainId


@dataclass
class ContractInfo:
    address: Address
    deployment_block: BlockNumber
    abi: list


def make_contracts(w3: web3.Web3, contracts_info: dict[str, ContractInfo]) -> dict[str, Contract]:
    return {
        name: w3.eth.contract(info.address, abi=info.abi) for name, info in contracts_info.items()
    }


def load_contract_abi(deployment_dir: Path, contract_name: str) -> list:
    with deployment_dir.joinpath(f"{contract_name}.json").open("rt") as f:
        data = json.load(f)
    return data["abi"]


DeploymentInfo = dict[ChainId, dict[str, ContractInfo]]


def load_deployment_info(deployment_dir: Path) -> DeploymentInfo:
    abis = {}
    deployment_info = {}
    with deployment_dir.joinpath("deployment.json").open("rt") as f:
        deployment = json.load(f)

    for chain_id, deployed_contracts in deployment["L2"].items():
        infos = {}
        for name, deployment_data in deployed_contracts.items():
            if name not in abis:
                abis[name] = load_contract_abi(deployment_dir, name)
            abi = abis[name]
            infos[name] = ContractInfo(
                address=deployment_data["address"],
                deployment_block=deployment_data["deployment_block"],
                abi=abi,
            )
        deployment_info[ChainId(int(chain_id))] = infos
    return deployment_info



import time
from copy import deepcopy
from unittest.mock import patch

import pytest
from hexbytes import HexBytes

from beamer.chain import process_claims
from beamer.events import FillHashInvalidated, HashInvalidated, InitiateL1InvalidationEvent
from beamer.state_machine import process_event
from beamer.tests.agent.unit.utils import (
    ACCOUNT,
    ADDRESS1,
    BLOCK_NUMBER,
    CLAIM_ID,
    TARGET_CHAIN_ID,
    TIMESTAMP,
    make_claim_challenged,
    make_claim_unchallenged,
    make_context,
    make_request,
)
from beamer.tests.agent.utils import make_address, make_tx_hash
from beamer.tests.constants import FILL_ID
from beamer.typing import ClaimId, FillId, Termination


def test_handle_fill_hash_invalidated():
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", FILL_ID, TIMESTAMP)
    context.requests.add(request.id, request)
    assert request.fill_id is not None

    claim_1 = make_claim_unchallenged(
        request=request, claimer=config.account.address, fill_id=request.fill_id
    )
    claim_1.start_challenge()
    claim_2 = make_claim_unchallenged(
        request=request,
        claim_id=ClaimId(claim_1.id + 1),
        claimer=config.account.address,
        fill_id=FillId(b"c0ffee"),
    )
    claim_2.start_challenge()
    context.claims.add(claim_1.id, claim_1)
    context.claims.add(claim_2.id, claim_2)

    fill_hash = request.fill_hash_with_fill_id(request.fill_id)
    event = FillHashInvalidated(
        chain_id=request.target_chain_id,
        tx_hash=HexBytes(""),
        fill_hash=fill_hash,
        block_number=BLOCK_NUMBER,
    )
    assert process_event(event, context) == (True, None)

    # Check that a matching event invalidates the claim
    assert claim_1.is_invalidated_l1_resolved  # pylint:disable=no-member

    # Check that a non-matching event does not invalidate the claim
    assert claim_2.is_claimer_winning  # pylint:disable=no-member


def test_handle_hash_invalidated():
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", FILL_ID, TIMESTAMP)
    context.requests.add(request.id, request)
    assert request.fill_id is not None

    claim = make_claim_unchallenged(
        request=request,
        claimer=config.account.address,
        fill_id=request.fill_id,
        stay_in_started_state=True,
    )
    context.claims.add(claim.id, claim)

    fill_hash = request.fill_hash_with_fill_id(request.fill_id)
    tx_hash = make_tx_hash()
    event = HashInvalidated(
        chain_id=request.target_chain_id,
        tx_hash=tx_hash,
        request_hash=request.request_hash,
        fill_id=request.fill_id,
        fill_hash=fill_hash,
        block_number=BLOCK_NUMBER,
    )
    assert process_event(event, context) == (True, None)

    # Check that the event changes the claim state
    assert claim.is_claimer_winning  # pylint:disable=no-member
    assert claim.invalidation_tx == tx_hash


@patch("beamer.chain._invalidate")
@pytest.mark.parametrize("fill_id", [FILL_ID, FillId(b"c0ffee")])
def test_maybe_invalidate_claim_wrong_fill_id(mocked_invalidate, fill_id):
    context, config = make_context()

    request = make_request(valid_until=int(time.time() * 2))
    request.fill(config.account.address, b"", FILL_ID, TIMESTAMP)
    context.requests.add(request.id, request)

    claim_event = make_claim_unchallenged(
        request=request,
        claimer=config.account.address,
        fill_id=fill_id,
    ).latest_claim_made
    process_event(claim_event, context)

    claim = context.claims.get(claim_event.claim_id)
    assert claim is not None

    assert claim.is_started
    process_claims(context)
    assert claim.is_claimer_winning
    # invalidate must only be called if the fill ids do not match
    assert mocked_invalidate.called == (fill_id != FILL_ID)


@patch("beamer.chain._invalidate")
def test_maybe_invalidate_claim_wrong_fill_id_but_in_back_off(mocked_invalidate):
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", FILL_ID, TIMESTAMP)
    context.requests.add(request.id, request)

    claim_event = make_claim_unchallenged(
        request=request,
        claimer=config.account.address,
        fill_id=FillId(b"c0ffee"),
    ).latest_claim_made
    assert request.fill_id != claim_event.fill_id
    process_event(claim_event, context)

    claim = context.claims.get(claim_event.claim_id)
    assert claim is not None
    claim.challenge_back_off_timestamp = int(time.time() + 100)

    assert claim.is_started
    process_claims(context)
    assert not claim.is_claimer_winning
    assert not mocked_invalidate.called


@patch("beamer.chain._invalidate")
def test_maybe_invalidate_claim_wrong_fill_id_but_timed_out(mocked_invalidate):
    context, config = make_context()

    request = make_request(valid_until=int(time.time() / 2))
    request.fill(config.account.address, b"", FILL_ID, TIMESTAMP)
    context.requests.add(request.id, request)

    claim_event = make_claim_unchallenged(
        request=request, claimer=config.account.address, fill_id=FillId(b"c0ffee")
    ).latest_claim_made
    assert request.fill_id != claim_event.fill_id
    process_event(claim_event, context)

    claim = context.claims.get(claim_event.claim_id)
    assert claim is not None

    assert claim.is_started
    process_claims(context)
    assert claim.is_claimer_winning
    assert mocked_invalidate.called


@patch("beamer.chain._withdraw")
@pytest.mark.parametrize(
    "test_data",
    [
        (ACCOUNT.address, ADDRESS1, False),
        (ADDRESS1, ACCOUNT.address, True),
        (ADDRESS1, make_address(), False),
    ],
)
def test_maybe_withdraw_after_invalidation(mocked_withdraw, test_data):
    context, _ = make_context()

    request = make_request()
    context.requests.add(request.id, request)

    claimer = test_data[0]
    challenger = test_data[1]
    should_withdraw = test_data[2]

    claim = make_claim_challenged(
        request=request,
        claimer=claimer,
        challenger=challenger,
        # Claim should not be withdrawable yet
        termination=Termination(TIMESTAMP + 1),
    )
    context.claims.add(claim.id, claim)
    claim.start_challenge(make_tx_hash(), TIMESTAMP)
    claim.l1_invalidate()
    assert claim.is_invalidated_l1_resolved  # pylint:disable=no-member

    assert not mocked_withdraw.called

    process_claims(context)

    assert should_withdraw == mocked_withdraw.called


@pytest.mark.parametrize("timestamp", [TIMESTAMP, int(time.time())])
def test_handle_generate_l1_invalidation_event(timestamp):
    context, config = make_context()

    request = make_request()
    context.requests.add(request.id, request)

    claim = make_claim_challenged(
        request=request,
        claimer=config.account.address,
    )
    claim.start_challenge(make_tx_hash(), timestamp)
    context.claims.add(claim.id, claim)

    event = deepcopy(claim.latest_claim_made)
    flag, events = process_event(event, context)

    assert flag

    assert events == [
        InitiateL1InvalidationEvent(
            chain_id=TARGET_CHAIN_ID,
            claim_id=CLAIM_ID,
        )
    ]



from typing import Optional

from eth_typing import ChecksumAddress as Address
from hexbytes import HexBytes
from statemachine import State, StateMachine
from web3.types import Timestamp, Wei

from beamer.events import ClaimMade
from beamer.models.request import Request
from beamer.typing import ClaimId, FillId, RequestId, Termination


class Claim(StateMachine):
    def __init__(
        self,
        claim_made: ClaimMade,
        challenge_back_off_timestamp: int,
    ) -> None:
        super().__init__()
        self._latest_claim_made = claim_made
        self._challenger_stakes: dict[Address, int] = {}
        self.challenge_back_off_timestamp = challenge_back_off_timestamp
        # transaction pending indicates whether a state output (transaction)
        # is pending in the chain network. If the corresponding event has
        # not arrived at the agent's EventProcessor yet, it will prevent
        # sending another transaction
        self.transaction_pending = False
        self.invalidation_tx: Optional[HexBytes] = None
        self.invalidation_timestamp: Optional[Timestamp] = None

    started = State("Started", initial=True)
    # Claimer is winning
    claimer_winning = State("ClaimerWinning")
    # Challenger is winning
    challenger_winning = State("ChallengerWinning")
    invalidated_l1_resolved = State("InvalidatedL1Resolved")
    ignored = State("Ignored")
    withdrawn = State("Withdrawn")

    start_challenge = (
        started.to(claimer_winning)
        | claimer_winning.to(claimer_winning)
        | challenger_winning.to(challenger_winning)
    )
    challenge = (
        claimer_winning.to(challenger_winning)
        | challenger_winning.to(claimer_winning)
        | ignored.to(ignored)
    )
    l1_invalidate = (
        claimer_winning.to(invalidated_l1_resolved)
        | challenger_winning.to(invalidated_l1_resolved)
        | invalidated_l1_resolved.to(invalidated_l1_resolved)
    )
    withdraw = (
        started.to(withdrawn)
        | claimer_winning.to(withdrawn)
        | challenger_winning.to(withdrawn)
        | ignored.to(withdrawn)
        | withdrawn.to(withdrawn)
    )
    ignore = claimer_winning.to(ignored) | ignored.to(ignored)

    @property
    def id(self) -> ClaimId:
        return self._latest_claim_made.claim_id

    @property
    def request_id(self) -> RequestId:
        return self._latest_claim_made.request_id

    @property
    def claimer(self) -> Address:
        return self._latest_claim_made.claimer

    @property
    def termination(self) -> Termination:
        return self._latest_claim_made.termination

    @property
    def fill_id(self) -> FillId:
        return self._latest_claim_made.fill_id

    @property
    def latest_claim_made(self) -> ClaimMade:
        return self._latest_claim_made

    def get_challenger_stake(self, challenger: Address) -> int:
        return self._challenger_stakes.get(challenger, 0)

    def add_challenger_stake(self, challenger: Address, amount: int) -> None:
        new_stake = amount + self.get_challenger_stake(challenger)
        self._challenger_stakes[challenger] = new_stake

    def valid_claim_for_request(self, request: Request) -> bool:
        claim_event = self._latest_claim_made
        if claim_event.request_id != request.id:
            return False
        if claim_event.claimer != request.filler:
            return False
        if claim_event.fill_id != request.fill_id:
            return False
        return True

    def get_winning_addresses(self) -> frozenset[Address]:
        if self._latest_claim_made.claimer_stake > self._latest_claim_made.challenger_stake_total:
            return frozenset({self.claimer})
        return frozenset(self._challenger_stakes.keys())

    def get_minimum_challenge_stake(self, initial_claim_stake: Wei) -> Wei:
        claimer_stake = self._latest_claim_made.claimer_stake
        challenger_stake = self._latest_claim_made.challenger_stake_total

        if challenger_stake > claimer_stake:
            return Wei(challenger_stake - claimer_stake + initial_claim_stake)
        else:
            return Wei(claimer_stake - challenger_stake + 1)

    def on_start_challenge(
        self, invalidation_tx: HexBytes = None, invalidation_timestamp: Timestamp = None
    ) -> None:
        self.invalidation_tx = invalidation_tx
        self.invalidation_timestamp = invalidation_timestamp

    def on_challenge(self, new_claim_made: ClaimMade) -> None:
        self._on_new_claim_made(new_claim_made)

    def on_withdraw(self) -> None:
        self.transaction_pending = False

    def on_ignore(self, new_claim_made: Optional[ClaimMade] = None) -> None:
        if new_claim_made is None:
            self._on_new_claim_made(self._latest_claim_made)

    def _on_new_claim_made(self, new_claim_made: ClaimMade) -> None:
        assert self._latest_claim_made.claim_id == new_claim_made.claim_id
        assert self._latest_claim_made.fill_id == new_claim_made.fill_id

        if new_claim_made.challenger_stake_total > new_claim_made.claimer_stake:
            new_challenger_stake = (
                new_claim_made.challenger_stake_total
                - self._latest_claim_made.challenger_stake_total
            )
            self.add_challenger_stake(new_claim_made.last_challenger, new_challenger_stake)

        self._latest_claim_made = new_claim_made
        self.transaction_pending = False

    def __repr__(self) -> str:
        state = self.current_state.identifier
        return f"<Claim id={self.id} state={state} request_id={self.request_id}>"



import brownie
from brownie import chain

from beamer.tests.constants import FILL_ID, RM_R_FIELD_LP_FEE, RM_R_FIELD_PROTOCOL_FEE
from beamer.tests.util import alloc_accounts, make_request, temp_fee_data

# Using this makes sure that we get nonzero fees when making requests.
_NONZERO_FEE_DATA = 14_000, 15_000, 5e18


def test_fee_split_works(deployer, request_manager, token, claim_stake, claim_period):
    requester, claimer = alloc_accounts(2)
    transfer_amount = 23_000_000

    request_manager.updateFeeData(*_NONZERO_FEE_DATA)
    request_id = make_request(
        request_manager, token, requester, requester, transfer_amount, fee_data="standard"
    )

    lp_fee = request_manager.lpFee(transfer_amount)
    assert lp_fee > 0

    protocol_fee = request_manager.protocolFee(transfer_amount)
    assert protocol_fee > 0

    assert lp_fee + protocol_fee == request_manager.totalFee(transfer_amount)

    # The request is not claimed yet, so no beamer fee has been collected yet
    assert request_manager.collectedProtocolFees(token) == 0
    assert request_manager.requests(request_id)[RM_R_FIELD_LP_FEE] == lp_fee
    assert request_manager.requests(request_id)[RM_R_FIELD_PROTOCOL_FEE] == protocol_fee

    claim_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim_tx.return_value

    # Update fees, which should not have any effect on the fee amounts that
    # were computed when the request was made.
    request_manager.updateFeeData(17e9, 145_000, 21_000, {"from": deployer})

    # Timetravel after claim period
    chain.mine(timedelta=claim_period)

    # Even if the requester calls withdraw, the funds go to the claimer
    withdraw_tx = request_manager.withdraw(claim_id, {"from": requester})
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert request_manager.collectedProtocolFees(token) == protocol_fee
    assert token.balanceOf(request_manager) == protocol_fee
    assert token.balanceOf(claimer) == transfer_amount + lp_fee


def test_protocol_fee_is_zero(request_manager):
    # For the time being, the protocol fee percentage should be zero.
    assert request_manager.protocolFeePPM() == 0
    assert request_manager.protocolFee(23_000_000) == 0


def test_protocol_fee_withdrawable_by_owner(
    deployer, request_manager, token, claim_stake, claim_period
):
    owner = deployer
    requester, claimer = alloc_accounts(2)
    amount = 23_000_000
    request_id = make_request(
        request_manager, token, requester, requester, amount, fee_data=_NONZERO_FEE_DATA
    )
    protocol_fee = request_manager.requests(request_id)[RM_R_FIELD_PROTOCOL_FEE]

    with brownie.reverts("Ownable: caller is not the owner"):
        request_manager.withdrawProtocolFees(token, requester, {"from": requester})

    with brownie.reverts("Protocol fee is zero"):
        request_manager.withdrawProtocolFees(token, owner, {"from": owner})

    claim_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim_tx.return_value

    chain.mine(timedelta=claim_period)

    with brownie.reverts("Protocol fee is zero"):
        request_manager.withdrawProtocolFees(token, owner, {"from": owner})

    request_manager.withdraw(claim_id, {"from": claimer})

    owner_token = token.balanceOf(owner)
    request_manager.withdrawProtocolFees(token.address, owner, {"from": owner})
    assert token.balanceOf(owner) == owner_token + protocol_fee


def test_fee_data_updatable_by_owner(deployer, request_manager):
    (requester,) = alloc_accounts(1)

    new_protocol_fee_ppm = 12_000
    new_lp_fee_ppm = 13_000
    new_min_lp_fee = 179e18

    with brownie.reverts("Ownable: caller is not the owner"):
        request_manager.updateFeeData(
            new_protocol_fee_ppm, new_lp_fee_ppm, new_min_lp_fee, {"from": requester}
        )

    request_manager.updateFeeData(
        new_protocol_fee_ppm, new_lp_fee_ppm, new_min_lp_fee, {"from": deployer}
    )
    assert request_manager.protocolFeePPM() == new_protocol_fee_ppm
    assert request_manager.lpFeePPM() == new_lp_fee_ppm
    assert request_manager.minLpFee() == new_min_lp_fee


def test_fee_reimbursed_on_expiration(request_manager, token):
    (requester,) = alloc_accounts(1)
    transfer_amount = 23_000_000
    validity_period = 60 * 5

    request_manager.updateFeeData(*_NONZERO_FEE_DATA)
    request_id = make_request(
        request_manager,
        token,
        requester,
        requester,
        transfer_amount,
        fee_data="standard",
        validity_period=validity_period,
    )

    total_fee = request_manager.totalFee(transfer_amount)
    assert total_fee > 0

    # Timetravel after validity period
    chain.mine(timedelta=validity_period)

    request_manager.withdrawExpiredRequest(request_id, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount + total_fee


def test_insufficient_lp_fee(request_manager, token):
    (requester,) = alloc_accounts(1)
    amount = 23_000_000
    validity_period = 60 * 5

    assert request_manager.lpFee(amount) > 0
    token.mint(requester, amount, {"from": requester})

    # Approve just the amount, ignoring the LP fee and the protocol fee.
    # This must cause createRequest to fail.
    token.approve(request_manager.address, amount, {"from": requester})

    with brownie.reverts("Insufficient allowance"):
        request_manager.createRequest(
            brownie.chain.id,
            token.address,
            token.address,
            requester,
            amount,
            validity_period,
            {"from": requester},
        )


def test_different_fees(request_manager, token, claim_period, claim_stake):
    (requester, claimer) = alloc_accounts(2)
    amount = 9e18
    fee_data_1 = 7_000, 8_000, 6e18
    fee_data_2 = 31_000, 4_000, 1

    token.mint(requester, amount * 10, {"from": requester})
    token.approve(request_manager.address, 2**256 - 1, {"from": requester})

    with temp_fee_data(request_manager, *fee_data_1):
        request_id_1 = make_request(
            request_manager, token, requester, requester, amount, fee_data="standard"
        )
        lp_fee_1 = request_manager.lpFee(amount)
        protocol_fee_1 = request_manager.protocolFee(amount)
        assert lp_fee_1 == 6e18
        assert protocol_fee_1 == 63e15

    with temp_fee_data(request_manager, *fee_data_2):
        request_id_2 = make_request(
            request_manager, token, requester, requester, amount, fee_data="standard"
        )
        lp_fee_2 = request_manager.lpFee(amount)
        protocol_fee_2 = request_manager.protocolFee(amount)
        assert lp_fee_2 == 36e15
        assert protocol_fee_2 == 279e15

    assert (
        token.balanceOf(request_manager)
        == amount * 2 + protocol_fee_1 + protocol_fee_2 + lp_fee_1 + lp_fee_2
    )
    claim_id_1 = request_manager.claimRequest(
        request_id_1, FILL_ID, {"from": claimer, "value": claim_stake}
    ).return_value
    claim_id_2 = request_manager.claimRequest(
        request_id_2, FILL_ID, {"from": claimer, "value": claim_stake}
    ).return_value

    chain.mine(timedelta=claim_period)

    withdraw_tx = request_manager.withdraw(claim_id_1, {"from": claimer})
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert request_manager.collectedProtocolFees(token) == protocol_fee_1
    assert token.balanceOf(request_manager) == amount + protocol_fee_1 + protocol_fee_2 + lp_fee_2
    assert token.balanceOf(claimer) == amount + lp_fee_1

    withdraw_tx = request_manager.withdraw(claim_id_2, {"from": claimer})
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert request_manager.collectedProtocolFees(token) == protocol_fee_1 + protocol_fee_2
    assert token.balanceOf(request_manager) == protocol_fee_1 + protocol_fee_2
    assert token.balanceOf(claimer) == amount * 2 + lp_fee_1 + lp_fee_2



import random
import string

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from hexbytes import HexBytes


def make_bytes(length: int) -> bytes:
    return bytes("".join(random.choice(string.printable) for _ in range(length)), encoding="utf-8")


def make_address() -> ChecksumAddress:
    return to_checksum_address(make_bytes(20))


def make_tx_hash() -> HexBytes:
    return HexBytes(make_bytes(32))



import pytest
from web3.types import Wei

from beamer.chain import process_claims
from beamer.tests.agent.unit.utils import (
    ADDRESS1,
    CLAIMER_STAKE,
    TIMESTAMP,
    make_claim_challenged,
    make_claim_unchallenged,
    make_context,
    make_request,
)
from beamer.tests.agent.utils import make_address
from beamer.tests.constants import FILL_ID
from beamer.typing import FillId, Termination


@pytest.mark.parametrize("fill_id", [FILL_ID, FillId(b"cafebabe")])
@pytest.mark.parametrize("filler", [ADDRESS1, make_address()])
def test_challenge_as_challenger(filler, fill_id):
    context, _ = make_context()

    request = make_request()
    context.requests.add(request.id, request)
    request.filler = ADDRESS1
    request.fill_id = FILL_ID
    claim = make_claim_unchallenged(
        request=request,
        claimer=filler,
        fill_id=fill_id,
        termination=Termination(TIMESTAMP + 1),
    )
    context.claims.add(claim.id, claim)

    assert not claim.transaction_pending
    process_claims(context)
    assert claim.valid_claim_for_request(request) != claim.transaction_pending


def test_challenge_as_claimer():
    context, _ = make_context()

    request = make_request()
    context.requests.add(request.id, request)
    request.filler = make_address()
    claim = make_claim_challenged(
        request=request,
        claimer=context.address,
        challenger=make_address(),
        challenger_stake=Wei(CLAIMER_STAKE + 1),
        termination=Termination(TIMESTAMP + 1),
    )
    context.claims.add(claim.id, claim)

    assert not claim.transaction_pending
    process_claims(context)
    assert claim.transaction_pending


@pytest.mark.parametrize("filler", [ADDRESS1, None])
def test_join_false_claim_challenge_only_when_unfilled(filler):

    context, _ = make_context()

    request = make_request()
    context.requests.add(request.id, request)
    request.filler = filler
    claim = make_claim_challenged(
        request=request,
        challenger=make_address(),
        termination=Termination(TIMESTAMP + 1),
    )
    context.claims.add(claim.id, claim)

    # Ensure that context is not part of challenge
    assert claim.claimer != context.address
    assert claim.get_challenger_stake(context.address) == 0
    assert claim.latest_claim_made.challenger_stake_total > 0

    process_claims(context)
    if request.filler is None:
        assert claim.transaction_pending
    else:
        assert not claim.transaction_pending



import json
import logging
import sys
from pathlib import Path
from typing import Any, List, Optional, TextIO, Union, cast

import requests
import structlog
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_utils import is_checksum_address, to_checksum_address
from web3 import HTTPProvider, Web3
from web3.contract import ContractConstructor, ContractFunction
from web3.exceptions import ContractLogicError
from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.middleware import construct_sign_and_send_raw_middleware, geth_poa_middleware
from web3.types import GasPriceStrategy, TxParams

from beamer.typing import URL, ChainId, ChecksumAddress


class TransactionFailed(Exception):
    def __repr__(self) -> str:
        return "TransactionFailed(%r)" % self.__cause__

    def __str__(self) -> str:
        return "transaction failed: %s" % self.cause()

    def cause(self) -> str:
        return str(self.__cause__)


def transact(
    func: Union[ContractConstructor, ContractFunction],
    timeout: float = 120,
    poll_latency: float = 0.1,
    **kwargs: Any,
) -> Any:
    try:
        txn_hash = func.transact(cast(Optional[TxParams], kwargs))
    except (ContractLogicError, requests.exceptions.RequestException) as exc:
        raise TransactionFailed() from exc

    return func.web3.eth.wait_for_transaction_receipt(
        txn_hash, timeout=timeout, poll_latency=poll_latency
    )


def setup_logging(log_level: str, log_json: bool) -> None:
    """Basic structlog setup"""

    logging.basicConfig(level=log_level, stream=sys.stdout, format="%(message)s")
    logging.logThreads = False

    logging.getLogger("web3").setLevel("INFO")
    logging.getLogger("urllib3").setLevel("INFO")

    shared_processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S.%f"),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
    ]

    if log_json:
        processors = shared_processors + [structlog.processors.JSONRenderer()]
    else:
        processors = shared_processors + [structlog.dev.ConsoleRenderer()]

    structlog.configure(
        processors=processors,  # type: ignore
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def account_from_keyfile(keyfile: Path, password: str) -> LocalAccount:
    with open(keyfile, "rt") as fp:
        privkey = Account.decrypt(json.load(fp), password)
    return cast(LocalAccount, Account.from_key(privkey))


def make_web3(
    url: URL, account: LocalAccount, gas_price_strategy: GasPriceStrategy = rpc_gas_price_strategy
) -> Web3:
    w3 = Web3(HTTPProvider(url, request_kwargs=dict(timeout=5)))
    w3.eth.set_gas_price_strategy(gas_price_strategy)
    # Add POA middleware for geth POA chains, no/op for other chains
    w3.middleware_onion.inject(geth_poa_middleware, layer=0)
    w3.middleware_onion.add(construct_sign_and_send_raw_middleware(account))
    w3.eth.default_account = account.address
    return w3


_Token = tuple[ChainId, ChecksumAddress]

# dictionary from base chain ID to deployed roll up IDs
SUPPORTED_CONNECTED_L2S = {
    # Mainnet
    1: frozenset({10, 42161, 288, 1088}),  # Optimism, Arbitrum, Boba, Metis
    # Rinkeby
    4: frozenset({421611, 28, 588}),  # Arbitrum, Boba, Metis
    # Goerli
    5: frozenset({420}),  # Optimism
    # Kovan
    42: frozenset({69}),  # Optimism
}


class TokenMatchChecker:
    def __init__(self, tokens: List[List[List[str]]]) -> None:
        # A mapping of tokens to equivalence classes. Each frozenset contains
        # tokens that are considered mutually equivalent.
        self._tokens: dict[_Token, frozenset[_Token]] = {}

        for token_mapping in tokens:
            equiv_class = frozenset(
                (ChainId(int(token[0])), to_checksum_address(token[1])) for token in token_mapping
            )
            l2_chain_ids = frozenset(chain_id for chain_id, _ in equiv_class)

            # check if equiv class contains chain ids from different base layers
            for connected_l2s in SUPPORTED_CONNECTED_L2S.values():
                intersection = connected_l2s.intersection(l2_chain_ids)
                if len(intersection) > 0:
                    msg = f"""All tokens' L2 chains must share the same base chain.
                     Please check {l2_chain_ids}"""
                    assert intersection == l2_chain_ids, msg

            for token in equiv_class:
                assert is_checksum_address(token[1])
                self._tokens[token] = equiv_class

    def is_valid_pair(
        self,
        source_chain_id: ChainId,
        source_token_address: ChecksumAddress,
        target_chain_id: ChainId,
        target_token_address: ChecksumAddress,
    ) -> bool:
        source_token = source_chain_id, source_token_address
        target_token = target_chain_id, target_token_address
        return target_token in self._tokens.get(source_token, frozenset())

    @staticmethod
    def from_file(f: TextIO) -> "TokenMatchChecker":
        tokens = json.load(f)
        return TokenMatchChecker(tokens)



import json
import os
import pathlib
import sys
import threading
import time
import traceback
from typing import Any, Callable

import structlog
from web3 import Web3
from web3.contract import Contract
from web3.types import Wei

from beamer.events import Event, EventFetcher
from beamer.models.claim import Claim
from beamer.models.request import Request
from beamer.state_machine import Context, process_event
from beamer.typing import BlockNumber, ChainId
from beamer.util import TransactionFailed, transact

log = structlog.get_logger(__name__)


def _load_ERC20_abi() -> list[Any]:
    path = pathlib.Path(__file__)
    path = path.parent.joinpath("data/abi/StandardToken.json")
    with path.open("rt") as fp:
        return json.load(fp)["abi"]


_ERC20_ABI = _load_ERC20_abi()

# The time we're waiting for our thread in stop(), in seconds.
# This is also the maximum time a call to stop() would block.
_STOP_TIMEOUT = 2

POLL_PERIOD_TESTNET = 1
POLL_PERIOD_MAINNET = 5


def _wrap_thread_func(func: Callable) -> Callable:
    def wrapper(*args, **kwargs):  # type: ignore
        try:
            return func(*args, **kwargs)
        except Exception:
            traceback.print_exception(*sys.exc_info())
            os._exit(1)
            # should never be reached
            return None

    return wrapper


class EventMonitor:
    def __init__(
        self,
        web3: Web3,
        contracts: tuple[Contract, ...],
        deployment_block: BlockNumber,
        on_new_events: Callable[[list[Event]], None],
        on_sync_done: Callable[[], None],
        poll_period: int,
    ):
        self._web3 = web3
        self._chain_id = ChainId(self._web3.eth.chain_id)
        self._contracts = contracts
        self._deployment_block = deployment_block
        self._stop = False
        self._on_new_events = on_new_events
        self._on_sync_done = on_sync_done
        self._poll_period = poll_period
        self._log = structlog.get_logger(type(self).__name__).bind(chain_id=self._chain_id)

        for contract in contracts:
            assert (
                self._chain_id == contract.web3.eth.chain_id
            ), f"Chain id mismatch for {contract}"

    def start(self) -> None:
        self._thread = threading.Thread(
            name=f"EventMonitor[cid={self._chain_id}]", target=_wrap_thread_func(self._thread_func)
        )
        self._thread.start()

    def stop(self) -> None:
        self._stop = True
        self._thread.join(_STOP_TIMEOUT)

    def _thread_func(self) -> None:
        self._log.info(
            "EventMonitor started",
            addresses=[c.address for c in self._contracts],
        )

        fetcher = EventFetcher(self._web3, self._contracts, self._deployment_block)
        events = fetcher.fetch()
        if events:
            self._on_new_events(events)
        self._on_sync_done()
        self._log.info("Sync done")
        while not self._stop:
            events = fetcher.fetch()
            if events:
                self._on_new_events(events)
            time.sleep(self._poll_period)
        self._log.info("EventMonitor stopped")


class EventProcessor:
    def __init__(self, context: Context):
        # This lock protects the following objects:
        #   - self._events
        #   - self._num_syncs_done
        self._lock = threading.Lock()
        self._have_new_events = threading.Event()
        self._events: list[Event] = []
        self._stop = False
        self._log = structlog.get_logger(type(self).__name__)
        # The number of times we synced with a chain:
        # 0 = we're still waiting for sync to complete for both chains
        # 1 = one of the chains was synced, waiting for the other one
        # 2 = both chains synced
        self._num_syncs_done = 0
        self._context = context

    @property
    def _synced(self) -> bool:
        with self._lock:
            return self._num_syncs_done == 2

    def mark_sync_done(self) -> None:
        with self._lock:
            assert self._num_syncs_done < 2
            self._num_syncs_done += 1

    def start(self) -> None:
        self._thread = threading.Thread(
            name="EventProcessor", target=_wrap_thread_func(self._thread_func)
        )
        self._thread.start()

    def stop(self) -> None:
        self._stop = True
        self._thread.join(_STOP_TIMEOUT)

    def add_events(self, events: list[Event]) -> None:
        with self._lock:
            self._events.extend(events)
            self._log.debug("New events", events=events)
        self._have_new_events.set()

    def _thread_func(self) -> None:
        self._log.info("EventProcessor started")

        chain_id = self._context.web3_l1.eth.chain_id
        poll_period = POLL_PERIOD_MAINNET if chain_id == 1 else POLL_PERIOD_TESTNET

        # Wait until all past events are fetched and in the queue
        # so the agent can sync to the current state
        while not self._synced and not self._stop:
            self._have_new_events.wait(poll_period)
            self._have_new_events.clear()

        while not self._stop:
            if self._have_new_events.wait(poll_period):
                self._have_new_events.clear()
            if self._events:
                self._process_events()

            process_requests(self._context)
            process_claims(self._context)

        self._log.info("EventProcessor stopped")

    def _process_events(self) -> None:
        iteration = 0
        created_events: list[Event] = []
        while True:
            with self._lock:
                events = self._events[:]

            unprocessed: list[Event] = []
            any_state_changed = False
            for event in events:
                state_changed, new_events = process_event(event, self._context)
                any_state_changed |= state_changed

                if new_events:
                    created_events.extend(new_events)
                if not state_changed:
                    unprocessed.append(event)

            # Return the unprocessed events to the event list.
            # Note that the event list might have been changed in the meantime
            # by one of the event monitors. Placing unprocessed events at the
            # back of the list, as opposed to the front, may avoid an extra
            # iteration over all events.
            with self._lock:
                del self._events[: len(events)]
                self._events.extend(unprocessed)

            self._log.debug(
                "Finished iteration",
                iteration=iteration,
                any_state_changed=any_state_changed,
                num_events=len(self._events),
            )
            iteration += 1
            if not any_state_changed:
                break

        # New events might be created by event handlers. If they would be
        # processed directly, there would be a chance of skipping certain states
        # in the request/claim state machines accidentally. So instead they are
        # collected and attached to the events list after the current batch of
        # events has been processed.
        with self._lock:
            self._events.extend(created_events)


def process_requests(context: Context) -> None:
    log.debug("Processing requests", num_requests=len(context.requests))

    to_remove = []
    for request in context.requests:
        log.debug("Processing request", request=request)

        if request.is_pending:
            fill_request(request, context)

        elif request.is_filled:
            claim_request(request, context)

        elif request.is_withdrawn or request.is_ignored:
            active_claims = any(claim.request_id == request.id for claim in context.claims)
            if not active_claims:
                log.debug("Removing request", request=request)
                to_remove.append(request.id)

    for request_id in to_remove:
        context.requests.remove(request_id)


def process_claims(context: Context) -> None:
    log.debug("Processing claims", num_claims=len(context.claims))

    to_remove = []
    for claim in context.claims:
        log.debug("Processing claim", claim=claim)

        request = context.requests.get(claim.request_id)
        # As per definition an invalid or expired request cannot be claimed
        # This gives us a chronological order. The agent should never garbage collect
        # a request which has active claims
        assert request is not None, "Active claim for non-existent request"

        if claim.is_ignored:
            continue

        if claim.is_started:
            # If the claim is not valid, we might need to send a non-fill-proof
            if not claim.valid_claim_for_request(request):
                maybe_invalidate(claim, context)
            else:
                claim.start_challenge()

            continue

        if claim.is_withdrawn:
            log.debug("Removing withdrawn claim", claim=claim)
            to_remove.append(claim.id)
            continue

        if claim.is_invalidated_l1_resolved:
            maybe_withdraw(claim, context)
            continue

        # Check if claim is an honest claim. Honest claims can be ignored.
        # This only counts for claims, where the agent is not the filler
        if claim.valid_claim_for_request(request) and request.filler != context.address:
            claim.ignore()
            continue

        if claim.transaction_pending:
            continue

        if claim.is_claimer_winning or claim.is_challenger_winning:
            maybe_withdraw(claim, context)
            maybe_challenge(claim, context)

    for claim_id in to_remove:
        context.claims.remove(claim_id)


def fill_request(request: Request, context: Context) -> None:
    block = context.latest_blocks[request.target_chain_id]
    if block["timestamp"] >= request.valid_until:
        log.info("Request expired, ignoring", request=request)
        request.ignore()
        return

    w3 = context.fill_manager.web3
    token = w3.eth.contract(abi=_ERC20_ABI, address=request.target_token_address)
    address = w3.eth.default_account
    balance = token.functions.balanceOf(address).call()
    if balance < request.amount:
        log.debug("Unable to fill request", balance=balance, request_amount=request.amount)
        return

    func = token.functions.approve(context.fill_manager.address, request.amount)
    try:
        transact(func)
    except TransactionFailed as exc:
        log.error("approve failed", request_id=request.id, cause=exc.cause())
        return

    func = context.fill_manager.functions.fillRequest(
        requestId=request.id,
        sourceChainId=request.source_chain_id,
        targetTokenAddress=request.target_token_address,
        targetReceiverAddress=request.target_address,
        amount=request.amount,
    )
    try:
        receipt = transact(func)
    except TransactionFailed as exc:
        log.error("fillRequest failed", request_id=request.id, cause=exc.cause())
        return

    request.try_to_fill()
    log.debug(
        "Filled request",
        request=request,
        txn_hash=receipt.transactionHash.hex(),
        token=token.functions.symbol().call(),
    )


def claim_request(request: Request, context: Context) -> None:
    if request.filler != context.address:
        return

    block = context.latest_blocks[request.source_chain_id]
    if block["timestamp"] >= request.valid_until:
        log.info("Request expired, ignoring", request=request)
        request.ignore()
        return

    stake = context.request_manager.functions.claimStake().call()

    func = context.request_manager.functions.claimRequest(request.id, request.fill_id)
    try:
        receipt = transact(func, value=stake)
    except TransactionFailed as exc:
        log.error(
            "claimRequest failed",
            request_id=request.id,
            fill_id=request.fill_id,
            cause=exc.cause(),
            stake=stake,
        )
        return

    request.try_to_claim()
    log.debug(
        "Claimed request",
        request=request,
        txn_hash=receipt.transactionHash.hex(),
    )


def maybe_challenge(claim: Claim, context: Context) -> bool:
    # We need to challenge if either of the following is true:
    # 1) The claim is dishonest AND claimer is winning
    # 2) Agent is claimer AND challenger is winning
    request = context.requests.get(claim.request_id)
    # As per definition an invalid or expired request cannot be claimed
    # This gives us a chronological order. The agent should never garbage collect
    # a request which has active claims
    assert request is not None, "Active claim for non-existent request"

    block = context.latest_blocks[request.source_chain_id]
    if block["timestamp"] >= claim.termination:
        return False

    if int(time.time()) < claim.challenge_back_off_timestamp:
        return False

    agent_winning = context.address in claim.get_winning_addresses()
    if agent_winning:
        return False

    own_challenge_stake = Wei(claim.get_challenger_stake(context.address))
    # With the condition above this means that the agent is currently losing
    agent_participating = claim.claimer == context.address or own_challenge_stake > 0

    if not agent_participating:
        if claim.is_challenger_winning:
            return False
        # Already challenged and has a filler, the agent won't challenge
        if request.filler is not None and claim.latest_claim_made.challenger_stake_total > 0:
            return False

    initial_claim_stake = context.request_manager.functions.claimStake().call()
    stake = claim.get_minimum_challenge_stake(initial_claim_stake)

    # TODO: have a central variable and proper L1 cost calculations
    l1_cost = Wei(initial_claim_stake + 10**15)

    if claim.latest_claim_made.challenger_stake_total > 0:
        stake = max(stake, Wei(l1_cost - own_challenge_stake))

    func = context.request_manager.functions.challengeClaim(claim.id)
    try:
        receipt = transact(func, value=stake)
    except TransactionFailed as exc:
        log.error("challengeClaim failed", claim=claim, cause=exc.cause(), stake=stake)
        return False

    claim.transaction_pending = True

    log.debug(
        "Challenged claim",
        claim=claim,
        txn_hash=receipt.transactionHash.hex(),
    )

    return True


def maybe_invalidate(claim: Claim, context: Context) -> None:
    request = context.requests.get(claim.request_id)
    # As per definition an invalid or expired request cannot be claimed
    # This gives us a chronological order. The agent should never garbage collect
    # a request which has active claims
    assert request is not None, "Active claim for non-existent request"

    timestamp = int(time.time())
    if timestamp < claim.challenge_back_off_timestamp:
        return

    _invalidate(request, claim, context)
    claim.start_challenge()


def maybe_withdraw(claim: Claim, context: Context) -> None:
    request = context.requests.get(claim.request_id)
    # As per definition an invalid or expired request cannot be claimed
    # This gives us a chronological order. The agent should never garbage collect
    # a request which has active claims
    assert request is not None, "Active claim for non-existent request"

    block = context.latest_blocks[request.source_chain_id]
    agent_is_claimer = claim.claimer == context.address
    agent_is_challenger = claim.get_challenger_stake(context.address) > 0

    # When request is L1 resolved, the termination isn't important
    if request.is_l1_resolved:
        # We claimed the request
        if (
            agent_is_claimer
            and request.l1_resolution_filler == context.address
            and request.l1_resolution_fill_id == claim.fill_id
        ):
            _withdraw(claim, context)

        # Claimer cheated
        if agent_is_challenger and (
            request.l1_resolution_filler != claim.claimer
            or request.l1_resolution_fill_id != claim.fill_id
        ):
            _withdraw(claim, context)

    # Claim has a non-fill proof and the agent is challenging
    elif claim.is_invalidated_l1_resolved and agent_is_challenger:
        _withdraw(claim, context)

    # Otherwise check that the challenge period is over
    elif block["timestamp"] >= claim.termination:
        winning_addresses = claim.get_winning_addresses()
        if context.address in winning_addresses:
            _withdraw(claim, context)


def _withdraw(claim: Claim, context: Context) -> None:
    func = context.request_manager.functions.withdraw(claim.id)
    try:
        receipt = transact(func)
    except TransactionFailed as exc:
        # Ignore the exception when the claim has been withdrawn already
        if "Claim already withdrawn" in exc.cause():
            claim.transaction_pending = True
            log.warning("Claim already withdrawn", claim=claim)
            return

        log.error("Withdraw failed", claim=claim, cause=exc.cause())
        return

    claim.transaction_pending = True
    log.debug("Withdrew", claim=claim.id, txn_hash=receipt.transactionHash.hex())


def _invalidate(request: Request, claim: Claim, context: Context) -> None:
    func = context.fill_manager.functions.invalidateFillHash(
        request.request_hash, claim.latest_claim_made.fill_id, request.source_chain_id
    )
    try:
        receipt = transact(func)
    except TransactionFailed as exc:
        log.error("Calling invalidateFillHash failed", claim=claim, cause=exc.cause())
        return

    log.debug("Invalidated fill hash", claim=claim.id, txn_hash=receipt.transactionHash.hex())



import os
import pathlib
from dataclasses import dataclass

import brownie
import eth_account
import pytest
from brownie import (
    FillManager,
    RequestManager,
    ResolutionRegistry,
    Resolver,
    TestL1Messenger,
    TestL2Messenger,
    TestProofSubmitter,
    accounts,
)

import beamer.metrics
from beamer.agent import Agent
from beamer.config import Config
from beamer.contracts import ContractInfo
from beamer.tests.util import alloc_accounts
from beamer.typing import BlockNumber


@dataclass(frozen=True)
class Contracts:
    resolver: Resolver
    fill_manager: FillManager
    request_manager: RequestManager
    l1_messenger: TestL1Messenger
    l2_messenger: TestL2Messenger
    proof_submitter: TestProofSubmitter
    resolution_registry: ResolutionRegistry


# brownie local account, to be used for fulfilling requests.
# The private key here corresponds to the 10th account ganache creates on
# startup.
_LOCAL_ACCOUNT = accounts.add("0x3ff6c8dfd3ab60a14f2a2d4650387f71fe736b519d990073e650092faaa621fa")


@pytest.fixture
def deployer():
    return alloc_accounts(1)[0]


# Make sure that the chain is reset after each test since brownie
# launches ganache only once for the entire test suite run.
@pytest.fixture(autouse=True)
def _reset_chain():
    yield
    brownie.chain.reset()


@pytest.fixture
def claim_stake():
    return 100


@pytest.fixture
def claim_period():
    return 100


@pytest.fixture
def finalization_time():
    return 200


@pytest.fixture
def challenge_period_extension():
    return 50


@pytest.fixture
def forward_state():
    return False


@pytest.fixture
def contracts(
    deployer,
    forward_state,
    claim_stake,
    claim_period,
    finalization_time,
    challenge_period_extension,
):
    # L1 contracts
    l1_messenger = deployer.deploy(TestL1Messenger)
    l1_messenger.setForwardState(forward_state)
    resolver = deployer.deploy(Resolver)

    # L2b contracts
    l2_messenger = deployer.deploy(TestL2Messenger)
    l2_messenger.setForwardState(forward_state)
    proof_submitter = deployer.deploy(TestProofSubmitter, l2_messenger.address)
    fill_manager = deployer.deploy(FillManager, resolver.address, proof_submitter.address)
    fill_manager.addAllowedLP(_LOCAL_ACCOUNT)

    # L2a contracts
    resolution_registry = deployer.deploy(ResolutionRegistry)
    request_manager = deployer.deploy(
        RequestManager,
        claim_stake,
        claim_period,
        challenge_period_extension,
        resolution_registry.address,
    )

    # Explicitly allow calls between contracts. The chain of trust:
    #
    # fill_manager -> proof_submitter -> L2 messenger -> L1 resolver ->
    # L1 messenger -> resolution registry
    l1_chain_id = l2_chain_id = brownie.chain.id

    proof_submitter.addCaller(l2_chain_id, fill_manager.address)
    l2_messenger.addCaller(l2_chain_id, proof_submitter.address)
    resolver.addCaller(l2_chain_id, l2_messenger.address, proof_submitter.address)
    l1_messenger.addCaller(l1_chain_id, resolver.address)
    resolution_registry.addCaller(l1_chain_id, l1_messenger.address, resolver.address)
    resolver.addRegistry(l2_chain_id, resolution_registry.address, l1_messenger.address)

    request_manager.setFinalizationTime(l2_chain_id, finalization_time)
    return Contracts(
        l1_messenger=l1_messenger,
        l2_messenger=l2_messenger,
        resolver=resolver,
        proof_submitter=proof_submitter,
        fill_manager=fill_manager,
        request_manager=request_manager,
        resolution_registry=resolution_registry,
    )


@pytest.fixture
def config(request_manager, fill_manager, resolution_registry, token):
    root = pathlib.Path(__file__).parents[2]
    token_match_file = root / "beamer/data/tokens.example.json"

    contracts_info = dict(
        RequestManager=ContractInfo(
            deployment_block=BlockNumber(1),
            address=request_manager.address,
            abi=request_manager.abi,
        ),
        FillManager=ContractInfo(
            deployment_block=BlockNumber(1), address=fill_manager.address, abi=fill_manager.abi
        ),
        ResolutionRegistry=ContractInfo(
            deployment_block=BlockNumber(1),
            address=resolution_registry.address,
            abi=resolution_registry.abi,
        ),
    )
    deployment_info = {brownie.chain.id: contracts_info}
    account = eth_account.Account.from_key(_LOCAL_ACCOUNT.private_key)
    token.mint(account.address, 300)
    url = brownie.web3.provider.endpoint_uri
    config = Config(
        l1_rpc_url=url,
        l2a_rpc_url=url,
        l2b_rpc_url=url,
        deployment_info=deployment_info,
        token_match_file=token_match_file,
        account=account,
        fill_wait_time=0,
        prometheus_metrics_port=None,
    )
    beamer.metrics.init(config)
    return config


@pytest.fixture
def allow_unlisted_pairs() -> bool:
    return True


@pytest.fixture
def set_allow_unlisted_pairs(allow_unlisted_pairs: bool) -> None:
    if allow_unlisted_pairs:
        os.environ["BEAMER_ALLOW_UNLISTED_PAIRS"] = "1"
    else:
        if os.environ.get("BEAMER_ALLOW_UNLISTED_PAIRS") is not None:
            del os.environ["BEAMER_ALLOW_UNLISTED_PAIRS"]


@pytest.fixture
def agent(config, set_allow_unlisted_pairs):  # pylint:disable=unused-argument
    agent = Agent(config)
    agent.start()
    yield agent
    agent.stop()


@pytest.fixture
def token(deployer, MintableToken):
    return deployer.deploy(MintableToken, int(1e18))


@pytest.fixture
def request_manager(contracts):
    return contracts.request_manager


@pytest.fixture
def test_cross_domain_messenger(contracts):
    return contracts.messenger


@pytest.fixture
def resolver(contracts):
    return contracts.resolver


@pytest.fixture
def resolution_registry(contracts):
    return contracts.resolution_registry


@pytest.fixture
def optimism_proof_submitter(contracts):
    return contracts.proof_submitter


@pytest.fixture
def fill_manager(contracts):
    return contracts.fill_manager



from typing import Optional

from eth_abi.packed import encode_abi_packed
from eth_typing import ChecksumAddress
from eth_utils import keccak, to_canonical_address
from hexbytes import HexBytes
from statemachine import State, StateMachine
from web3.types import Timestamp

from beamer.typing import ChainId, FillHash, FillId, RequestHash, RequestId, TokenAmount


class Request(StateMachine):
    def __init__(
        self,
        request_id: RequestId,
        source_chain_id: ChainId,
        target_chain_id: ChainId,
        source_token_address: ChecksumAddress,
        target_token_address: ChecksumAddress,
        target_address: ChecksumAddress,
        amount: TokenAmount,
        valid_until: int,
    ) -> None:
        super().__init__()
        self.id = request_id
        self.source_chain_id = source_chain_id
        self.target_chain_id = target_chain_id
        self.source_token_address = source_token_address
        self.target_token_address = target_token_address
        self.target_address = target_address
        self.amount = amount
        self.valid_until = valid_until
        self.filler: Optional[ChecksumAddress] = None
        self.fill_tx: Optional[HexBytes] = None
        self.fill_timestamp: Optional[Timestamp] = None
        self.fill_id: Optional[FillId] = None
        self.l1_resolution_filler: Optional[ChecksumAddress] = None
        self.l1_resolution_fill_id: Optional[FillId] = None

    pending = State("Pending", initial=True)
    filled = State("Filled")
    claimed = State("Claimed")
    l1_resolved = State("L1Resolved")
    withdrawn = State("Withdrawn")
    ignored = State("Ignored")

    fill = pending.to(filled) | filled.to(filled) | ignored.to(filled) | claimed.to(claimed)
    try_to_fill = pending.to(filled)
    try_to_claim = filled.to(claimed) | pending.to(claimed) | claimed.to(claimed)
    l1_resolve = filled.to(l1_resolved) | claimed.to(l1_resolved) | l1_resolved.to(l1_resolved)
    withdraw = (
        claimed.to(withdrawn)
        | filled.to(withdrawn)
        | l1_resolved.to(withdrawn)
        | ignored.to(withdrawn)
    )
    ignore = pending.to(ignored) | filled.to(ignored)

    def on_fill(
        self,
        filler: ChecksumAddress,
        fill_tx: HexBytes,
        fill_id: FillId,
        fill_timestamp: Timestamp,
    ) -> None:
        self.filler = filler
        self.fill_tx = fill_tx
        self.fill_id = fill_id
        self.fill_timestamp = fill_timestamp

    def on_l1_resolve(
        self, l1_filler: Optional[ChecksumAddress] = None, l1_fill_id: Optional[FillId] = None
    ) -> None:
        self.l1_resolution_filler = l1_filler
        self.l1_resolution_fill_id = l1_fill_id

    @property
    def request_hash(self) -> RequestHash:
        return RequestHash(
            keccak(
                encode_abi_packed(
                    ["uint256", "uint256", "uint256", "address", "address", "uint256"],
                    [
                        self.id,
                        self.source_chain_id,
                        self.target_chain_id,
                        to_canonical_address(self.target_token_address),
                        to_canonical_address(self.target_address),
                        self.amount,
                    ],
                )
            )
        )

    def fill_hash_with_fill_id(self, fill_id: FillId) -> FillHash:
        return FillHash(
            keccak(
                encode_abi_packed(
                    ["bytes32", "bytes32"],
                    [
                        self.request_hash,
                        fill_id,
                    ],
                )
            )
        )

    def __repr__(self) -> str:
        state = self.current_state.identifier
        return f"<Request id={self.id} state={state} filler={self.filler}>"



import brownie
import pytest
from brownie.convert.datatypes import HexString
from eth_abi.packed import encode_abi_packed
from eth_utils import keccak
from web3.constants import ADDRESS_ZERO

from beamer.tests.constants import FILL_ID_EMPTY
from beamer.tests.util import alloc_accounts, create_request_hash


@pytest.mark.parametrize("amount", [100, 99, 101])
@pytest.mark.parametrize("forward_state", [True])
def test_l1_resolution_correct_hash(fill_manager, deployer, resolution_registry, token, amount):
    requested_amount = 100
    request_id = 23
    filler, receiver = alloc_accounts(2)
    chain_id = brownie.web3.eth.chain_id

    fill_manager.addAllowedLP(filler, {"from": deployer})

    token.mint(filler, amount, {"from": filler})
    token.approve(fill_manager.address, amount, {"from": filler})

    fill_tx = fill_manager.fillRequest(
        request_id, chain_id, token.address, receiver, amount, {"from": filler}
    )
    fill_id = fill_tx.return_value

    request_hash = create_request_hash(
        request_id,
        chain_id,
        chain_id,
        token.address,
        receiver.address,
        requested_amount,
    )

    expected_address = ADDRESS_ZERO
    expected_fill_id = FILL_ID_EMPTY

    if amount == requested_amount:
        expected_address = filler
        expected_fill_id = fill_id

    resolved_filler, resolved_fill_id = resolution_registry.fillers(request_hash)

    assert resolved_filler == expected_address
    assert resolved_fill_id == HexString(expected_fill_id, "bytes32")


@pytest.mark.parametrize("forward_state", [True])
def test_l1_non_fill_proof(fill_manager, resolution_registry):
    request_hash = "1234" + "00" * 30
    fill_id = "5678" + "00" * 30
    chain_id = brownie.web3.eth.chain_id

    fill_manager.invalidateFillHash(request_hash, fill_id, chain_id)

    fill_hash = keccak(
        encode_abi_packed(
            ["bytes32", "bytes32"],
            [
                bytes.fromhex(request_hash),
                bytes.fromhex(fill_id),
            ],
        )
    )

    assert resolution_registry.invalidFillHashes(fill_hash)


def test_restricted_calls(contracts, resolver):
    """Test that important contract calls cannot be invoked by a random caller."""
    (caller,) = alloc_accounts(1)

    # fill_manager -> proof_submitter -> messenger1 -> L1 resolver ->
    # messenger2 -> resolution registry
    with brownie.reverts("RestrictedCalls: unknown caller"):
        contracts.proof_submitter.submitProof(
            resolver, brownie.chain.id, 0, caller, {"from": caller}
        )

    with brownie.reverts("XRestrictedCalls: unknown caller"):
        contracts.resolver.resolve(
            0, 0, brownie.chain.id, brownie.chain.id, caller, {"from": caller}
        )

    with brownie.reverts("XRestrictedCalls: unknown caller"):
        contracts.resolver.resolveNonFill(
            0, 0, brownie.chain.id, brownie.chain.id, {"from": caller}
        )

    with brownie.reverts("XRestrictedCalls: unknown caller"):
        contracts.resolution_registry.resolveRequest(0, 0, 0, caller, {"from": caller})

    with brownie.reverts("XRestrictedCalls: unknown caller"):
        contracts.resolution_registry.invalidateFillHash(0, 0, 0, {"from": caller})



import threading
from typing import Any, Generator, Generic, Optional, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class Tracker(Generic[K, V]):
    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._map: dict[K, V] = {}

    def add(self, key: K, value: V) -> None:
        with self._lock:
            self._map[key] = value

    def remove(self, key: K) -> None:
        with self._lock:
            del self._map[key]

    def __contains__(self, key: K) -> bool:
        with self._lock:
            return key in self._map

    def get(self, key: K) -> Optional[V]:
        return self._map.get(key)

    def __iter__(self) -> Any:
        def locked_iter() -> Generator:
            with self._lock:
                it = iter(self._map.values())
                while True:
                    try:
                        yield next(it)
                    except StopIteration:
                        return

        return locked_iter()

    def __len__(self) -> int:
        with self._lock:
            return len(self._map)



import contextlib
import threading
from dataclasses import dataclass
from typing import Any, Generator

import structlog
from prometheus_client import Counter, Info, start_http_server

log = structlog.get_logger(__name__)


def init(config: Any) -> None:
    global _DATA

    if _DATA is not None:
        return

    info = Info("agent_info", "Agent information")
    info.info(
        dict(
            account=config.account.address,
            l2a_rpc_url=config.l2a_rpc_url,
            l2b_rpc_url=config.l2b_rpc_url,
        )
    )
    requests_filled = Counter(
        "requests_filled",
        "Number of requests filled on the target rollup, regardless of filler",
    )
    requests_filled_by_agent = Counter(
        "requests_filled_by_agent", "Number of requests the agent filled on the target rollup"
    )
    requests_created = Counter(
        "requests_created", "Number of requests created on the source rollup"
    )

    _DATA = _Data(
        info=info,
        requests_filled=requests_filled,
        requests_filled_by_agent=requests_filled_by_agent,
        requests_created=requests_created,
    )
    if config.prometheus_metrics_port is not None:
        log.info("Serving Prometheus metrics", port=config.prometheus_metrics_port)
        start_http_server(config.prometheus_metrics_port)


@dataclass
class _Data:
    info: Info
    requests_filled: Counter
    requests_filled_by_agent: Counter
    requests_created: Counter


_DATA: _Data = None  # type:ignore
_DATA_LOCK = threading.Lock()


@contextlib.contextmanager
def update() -> Generator[_Data, None, None]:
    with _DATA_LOCK:
        yield _DATA



import brownie
from brownie import web3
from web3.constants import ADDRESS_ZERO

from beamer.tests.agent.utils import make_address
from beamer.tests.constants import FILL_ID
from beamer.tests.util import create_fill_hash_from_request_hash, create_request_hash


def test_invalidation_before_and_after_resolution(contracts, token, resolution_registry):
    contracts.l1_messenger.setLastSender(contracts.resolver.address)

    address = make_address()
    request_id = 1
    chain_id = web3.eth.chain_id

    request_hash = create_request_hash(
        request_id,
        chain_id,
        chain_id,
        token.address,
        address,
        0,
    )

    fill_hash = create_fill_hash_from_request_hash(
        request_hash,
        FILL_ID,
    )

    assert not resolution_registry.invalidFillHashes(fill_hash)
    resolution_registry.invalidateFillHash(
        request_hash, FILL_ID, chain_id, {"from": contracts.l1_messenger}
    )

    # Fill hash must be invalidated
    assert resolution_registry.invalidFillHashes(fill_hash)

    assert resolution_registry.fillers(request_hash)[0] == ADDRESS_ZERO

    resolution_registry.resolveRequest(
        request_hash, FILL_ID, chain_id, address, {"from": contracts.l1_messenger}
    )

    # Resolution validates fill hash again
    assert resolution_registry.fillers(request_hash)[0] == address
    assert not resolution_registry.invalidFillHashes(fill_hash)

    # Invalidation of a resolved request should fail
    with brownie.reverts("Cannot invalidate resolved fillHashes"):
        resolution_registry.invalidateFillHash(
            request_hash, FILL_ID, chain_id, {"from": contracts.l1_messenger}
        )



import time
from unittest.mock import patch

import brownie
import pytest
from brownie import ZERO_ADDRESS
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from web3.types import Timestamp, Wei

from beamer.agent import Agent
from beamer.chain import maybe_challenge
from beamer.events import ClaimMade, RequestFilled
from beamer.models.claim import Claim
from beamer.models.request import Request
from beamer.tests.agent.unit.utils import BLOCK_NUMBER
from beamer.tests.util import HTTPProxy, Sleeper, Timeout, alloc_accounts, make_request
from beamer.typing import ChainId, ClaimId, FillId, RequestId, Termination, TokenAmount


def _get_delay(request_data):
    params = request_data["params"][0]
    to_block = int(params["toBlock"], 16)
    from_block = int(params["fromBlock"], 16)
    num_blocks = to_block - from_block + 1
    if num_blocks <= 200:
        return 0
    return 5


def test_read_timeout(config):
    brownie.chain.mine(200)

    proxy_l2a = HTTPProxy(config.l2a_rpc_url)
    proxy_l2a.delay_rpc({"eth_getLogs": _get_delay})
    proxy_l2a.start()

    proxy_l2b = HTTPProxy(config.l2b_rpc_url)
    proxy_l2b.delay_rpc({"eth_getLogs": _get_delay})
    proxy_l2b.start()

    config.l2a_rpc_url = "http://%s:%s" % (
        proxy_l2a.server_address[0],
        proxy_l2a.server_address[1],
    )
    config.l2b_rpc_url = "http://%s:%s" % (
        proxy_l2b.server_address[0],
        proxy_l2b.server_address[1],
    )

    agent = Agent(config)
    agent.start()
    time.sleep(60)
    agent.stop()
    proxy_l2a.stop()
    proxy_l2b.stop()


# This test delays the processing of eth_sendRawTransaction so that an agent
# sending a transaction gets a timeout. Previously, that would result in an
# exception that would propagate upwards within the thread, causing the agent
# to exit. With a proper fix the agent should catch the exception, continue
# running and make a clean exit after our call agent.stop().
def test_read_timeout_send_transaction(request_manager, token, config):
    delay_period = 6
    proxy_l2a = HTTPProxy(config.l2a_rpc_url)
    proxy_l2a.delay_rpc({"eth_sendRawTransaction": delay_period})
    proxy_l2a.start()

    config.l2a_rpc_url = "http://%s:%s" % (
        proxy_l2a.server_address[0],
        proxy_l2a.server_address[1],
    )

    requester, target = alloc_accounts(2)
    make_request(request_manager, token, requester, target, 1, validity_period=3600)

    agent = Agent(config)
    agent.start()
    time.sleep(delay_period)
    agent.stop()
    proxy_l2a.stop()


def test_challenge_own_claim(config, request_manager, token):
    agent = Agent(config)
    agent_address = to_checksum_address(agent.address)
    claim_stake = request_manager.claimStake()
    request = Request(
        RequestId(100),
        brownie.chain.id,
        brownie.chain.id,
        token.address,
        token.address,
        agent_address,
        TokenAmount(1),
        Termination(1700000000),
    )

    claim = Claim(
        ClaimMade(
            chain_id=brownie.chain.id,
            tx_hash=HexBytes(b""),
            claim_id=ClaimId(1),
            request_id=request.id,
            fill_id=FillId(b"1"),
            claimer=agent_address,
            claimer_stake=claim_stake,
            last_challenger=ZERO_ADDRESS,
            challenger_stake_total=Wei(0),
            termination=Termination(1700000000),
            block_number=BLOCK_NUMBER,
        ),
        int(time.time()),
    )
    # Add context so that maybe_challenge verifies that the claim is not expired
    agent.context.requests.add(request.id, request)
    agent.context.latest_blocks[brownie.chain.id] = {"timestamp": Timestamp(0)}

    assert not maybe_challenge(claim, agent.context), "Tried to challenge own claim"


@pytest.mark.parametrize("allow_unlisted_pairs", (True, False))
def test_fill_and_claim(request_manager, token, agent, allow_unlisted_pairs):
    requester, target = alloc_accounts(2)
    request_id = make_request(request_manager, token, requester, target, 1)

    try:
        with Sleeper(5) as sleeper:
            while (request := agent.context.requests.get(request_id)) is None:
                sleeper.sleep(0.1)
    except Timeout:
        pass

    if not allow_unlisted_pairs:
        assert request is None
        return
    else:
        assert request.id == request_id

    with Sleeper(5) as sleeper:
        while not request.is_claimed:
            sleeper.sleep(0.1)

    found_claims: list[Claim] = []
    with Sleeper(5) as sleeper:
        while not found_claims:
            for claim in agent.context.claims:
                if claim.request_id == request_id:
                    found_claims.append(claim)
            sleeper.sleep(0.1)

    assert len(found_claims) == 1
    claim = found_claims[0]
    assert claim.claimer == agent.address


def test_withdraw(request_manager, token, agent):
    requester, target = alloc_accounts(2)
    request_id = make_request(request_manager, token, requester, target, 1)

    with Sleeper(10) as sleeper:
        while (request := agent.context.requests.get(request_id)) is None:
            sleeper.sleep(0.1)

        while not request.is_claimed:
            sleeper.sleep(0.1)

    claim_period = request_manager.claimPeriod()
    brownie.chain.mine(timedelta=claim_period)

    with Sleeper(5) as sleeper:
        while not request.is_withdrawn:
            sleeper.sleep(0.1)


def test_expired_request_is_ignored(request_manager, token, agent):
    requester, target = alloc_accounts(2)
    validity_period = request_manager.MIN_VALIDITY_PERIOD()
    # make the request amount high enough that the agent cannot fill it
    amount = token.balanceOf(agent.address) + 1
    request_id = make_request(
        request_manager,
        token,
        requester,
        target,
        amount,
        validity_period=validity_period,
    )

    brownie.chain.mine(timedelta=validity_period / 2)
    with Sleeper(1) as sleeper:
        while (request := agent.context.requests.get(request_id)) is None:
            sleeper.sleep(0.1)

    assert request.is_pending

    brownie.chain.mine(timedelta=validity_period / 2 + 1)
    with Sleeper(2) as sleeper:
        while not request.is_ignored:
            sleeper.sleep(0.1)


# Disable filling of requests in the agent
# TODO: Find a better way to do this
@patch("beamer.chain.fill_request")
def test_agent_ignores_invalid_fill(_, request_manager, token, agent: Agent):
    requester, target, filler = alloc_accounts(3)
    validity_period = request_manager.MIN_VALIDITY_PERIOD()
    chain_id = ChainId(brownie.chain.id)
    amount = token.balanceOf(agent.address)

    request_id = make_request(
        request_manager,
        token,
        requester,
        target,
        amount,
        validity_period=validity_period,
    )

    with Sleeper(1) as sleeper:
        while (request := agent.context.requests.get(request_id)) is None:
            sleeper.sleep(0.1)

    event_processor = agent._event_processor

    # Test wrong amount
    event_processor.add_events(
        [
            RequestFilled(
                chain_id=chain_id,
                tx_hash=HexBytes(b""),
                request_id=RequestId(request_id),
                fill_id=FillId(b"1"),
                source_chain_id=chain_id,
                target_token_address=token,
                filler=filler,
                amount=amount - 1,
                block_number=brownie.web3.eth.block_number,
            ),
        ]
    )
    time.sleep(1)
    assert not request.is_filled

    # Test wrong `source_chain_id`
    event_processor.add_events(
        [
            RequestFilled(
                chain_id=chain_id,
                tx_hash=HexBytes(b""),
                request_id=RequestId(request_id),
                fill_id=FillId(b"1"),
                source_chain_id=ChainId(chain_id + 1),
                target_token_address=token,
                filler=filler,
                amount=amount,
                block_number=brownie.web3.eth.block_number,
            ),
        ]
    )
    time.sleep(1)
    assert not request.is_filled

    # Test wrong `target_token_address`
    event_processor.add_events(
        [
            RequestFilled(
                chain_id=chain_id,
                tx_hash=HexBytes(b""),
                request_id=RequestId(request_id),
                fill_id=FillId(b"1"),
                source_chain_id=chain_id,
                target_token_address=filler,
                filler=filler,
                amount=amount,
                block_number=brownie.web3.eth.block_number,
            ),
        ]
    )
    time.sleep(1)
    assert not request.is_filled

    # Test correct event
    event_processor.add_events(
        [
            RequestFilled(
                chain_id=chain_id,
                tx_hash=HexBytes(b""),
                request_id=RequestId(request_id),
                fill_id=FillId(b"1"),
                source_chain_id=chain_id,
                target_token_address=token,
                filler=filler,
                amount=amount,
                block_number=brownie.web3.eth.block_number,
            ),
        ]
    )
    with Sleeper(1) as sleeper:
        while not request.is_filled:
            sleeper.sleep(0.1)



import contextlib
import json
import threading
import time
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Any, List, Optional

import brownie
import requests
import web3
from eth_abi.packed import encode_abi_packed
from eth_utils import keccak, to_canonical_address

from beamer.typing import RequestId


def _alloc_account():
    address = brownie.web3.geth.personal.new_account("")
    account = brownie.accounts.at(address)
    assert brownie.web3.geth.personal.unlock_account(address, "", 0)
    # transfer 1 ETH to the newly created account
    brownie.accounts[0].transfer(account, brownie.web3.toWei(1, "ether"))
    return account


def alloc_accounts(n):
    return tuple(_alloc_account() for _ in range(n))


class Timeout(Exception):
    pass


class Sleeper:
    def __init__(self, timeout: float) -> None:
        self.timeout = timeout

    def __enter__(self):
        self._end = time.time() + self.timeout
        return self

    def __exit__(self, type_, value, traceback):
        pass

    def sleep(self, interval: float) -> None:
        if time.time() > self._end:
            raise Timeout()
        time.sleep(interval)


class _HTTPRequestHandler(BaseHTTPRequestHandler):
    def do_POST(self):
        self.server: HTTPProxy  # make mypy happy

        content_len = int(self.headers.get("Content-Length"))
        post_body = self.rfile.read(content_len)
        url = self.server.target_address
        try:
            response = requests.post(url, data=post_body)
        except requests.exceptions.ConnectionError:
            self.send_response_only(HTTPStatus.SERVICE_UNAVAILABLE)
            return

        data = json.loads(post_body)
        method = data["method"]
        delay = self.server.call_delays.get(method)
        if delay is not None:
            if callable(delay):
                delay = delay(data)
            time.sleep(delay)

        try:
            self.send_response(response.status_code)
            self.end_headers()
            self.wfile.write(response.content)
            self.wfile.flush()
        except ConnectionError:
            pass


class HTTPProxy(HTTPServer):
    def __init__(self, target_address):
        super().__init__(("", 0), _HTTPRequestHandler)
        self.target_address = target_address
        self.call_delays = {}

    def start(self) -> None:
        self._thread = threading.Thread(target=self.serve_forever)
        self._thread.start()

    def stop(self) -> None:
        self.shutdown()
        self._thread.join()

    def delay_rpc(self, call_delays):
        self.call_delays = call_delays


class EventCollector:
    def __init__(self, contract: web3.contract.Contract, event: str) -> None:
        self._address = contract.address
        contract = brownie.web3.eth.contract(address=contract.address, abi=contract.abi)
        self._event = getattr(contract.events, event)()
        self._events: List[Any] = []
        self._from_block = 0

    def next_event(self, wait_time: float = 10) -> Optional[Any]:
        """Return the next event. If no event comes within `wait_time` seconds,
        return None."""
        with Sleeper(wait_time) as sleeper:
            while not self._events:
                self._fetch_events()
                try:
                    sleeper.sleep(0.1)
                except Timeout:
                    return None
        return self._events.pop(0).args

    def _fetch_events(self) -> None:
        to_block = brownie.chain.height
        if to_block < self._from_block:
            return
        params = dict(fromBlock=self._from_block, toBlock=to_block, address=self._address)
        logs = brownie.web3.eth.get_logs(params)

        for log in logs:
            try:
                self._events.append(self._event.processLog(log))
            except web3.exceptions.MismatchedABI:
                pass
        self._from_block = to_block + 1


@contextlib.contextmanager
def earnings(w3, account):
    address = account.address
    balance_before = w3.eth.get_balance(address)

    block_before = w3.eth.block_number

    def calculate_gas_spending():
        total = 0
        for block_number in range(block_before + 1, brownie.chain.height + 1):
            block = brownie.web3.eth.get_block(block_number)
            for tx_hash in block.transactions:
                tx = brownie.web3.eth.get_transaction(tx_hash)
                tx_receipt = brownie.web3.eth.get_transaction_receipt(tx_hash)
                if tx["from"] == address:
                    total += tx.gasPrice * tx_receipt.gasUsed

        return total

    yield lambda: w3.eth.get_balance(address) + calculate_gas_spending() - balance_before


def get_fee_data(request_manager):
    return (
        request_manager.protocolFeePPM(),
        request_manager.lpFeePPM(),
        request_manager.minLpFee(),
    )


@contextlib.contextmanager
def temp_fee_data(request_manager, protocol_fee_ppm, lp_fee_ppm, min_lp_fee):
    old_fee_data = get_fee_data(request_manager)
    request_manager.updateFeeData(protocol_fee_ppm, lp_fee_ppm, min_lp_fee)
    yield
    request_manager.updateFeeData(*old_fee_data)


def make_request(
    request_manager,
    token,
    requester,
    target_address,
    amount,
    validity_period=3600,
    target_chain_id=None,
    fee_data=(0, 0, 0),
) -> RequestId:
    fees_context: Any  # make mypy happy
    if fee_data == "standard":
        fees_context = contextlib.nullcontext()
    else:
        fees_context = temp_fee_data(request_manager, *fee_data)

    with fees_context:
        total_token_amount = amount + request_manager.totalFee(amount)
        if token.balanceOf(requester) < total_token_amount:
            token.mint(requester, total_token_amount, {"from": requester})

        token.approve(request_manager.address, total_token_amount, {"from": requester})

        if target_chain_id is None:
            target_chain_id = brownie.chain.id

        request_tx = request_manager.createRequest(
            target_chain_id,
            token.address,
            token.address,
            target_address,
            amount,
            validity_period,
            {"from": requester},
        )
    return request_tx.return_value


def create_request_hash(
    request_id, source_chain_id, target_chain_id, target_token_address, receiver_address, amount
):
    return keccak(
        encode_abi_packed(
            ["uint256", "uint256", "uint256", "address", "address", "uint256"],
            [
                request_id,
                source_chain_id,
                target_chain_id,
                to_canonical_address(target_token_address),
                to_canonical_address(receiver_address),
                amount,
            ],
        )
    )


def create_fill_hash(
    request_id,
    source_chain_id,
    target_chain_id,
    target_token_address,
    receiver_address,
    amount,
    fill_id,
):
    return create_fill_hash_from_request_hash(
        create_request_hash(
            request_id,
            source_chain_id,
            target_chain_id,
            target_token_address,
            receiver_address,
            amount,
        ),
        fill_id,
    )


def create_fill_hash_from_request_hash(
    request_hash,
    fill_id,
):
    return keccak(
        encode_abi_packed(
            ["bytes32", "bytes32"],
            [
                request_hash,
                fill_id,
            ],
        )
    )



import time

import pytest
from web3.types import Wei

from beamer.events import InitiateL1InvalidationEvent, InitiateL1ResolutionEvent
from beamer.state_machine import process_event
from beamer.tests.agent.unit.utils import (
    CLAIM_ID,
    REQUEST_ID,
    TARGET_CHAIN_ID,
    TIMESTAMP,
    make_claim_challenged,
    make_context,
    make_request,
)
from beamer.tests.agent.utils import make_tx_hash


@pytest.mark.parametrize("timestamp", [TIMESTAMP, int(time.time())])
def test_handle_initiate_l1_resolution(timestamp):
    context, config = make_context()
    context.finalization_times[TARGET_CHAIN_ID] = 1_000

    request = make_request()
    context.requests.add(request.id, request)

    event = InitiateL1ResolutionEvent(
        chain_id=TARGET_CHAIN_ID,
        request_id=REQUEST_ID,
        claim_id=CLAIM_ID,
    )

    # Without a claim, we simply drop the event
    assert process_event(event, context) == (True, None)

    claim = make_claim_challenged(request, claimer=config.account.address)
    context.claims.add(claim.id, claim)

    # Must only be called if request is filled
    with pytest.raises(AssertionError, match="Request not yet filled"):
        process_event(event, context)

    # Check that task is added to resolution pool
    context.web3_l1.eth.gas_price = Wei(1)  # type: ignore
    request.fill(config.account.address, b"", b"", timestamp)
    request.try_to_claim()

    if timestamp == TIMESTAMP:
        assert process_event(event, context) == (True, None)
        assert context.task_pool.submit.called  # type: ignore  # pylint:disable=no-member
    else:
        assert process_event(event, context) == (False, None)
        assert not context.task_pool.submit.called  # type: ignore  # pylint:disable=no-member


@pytest.mark.parametrize("timestamp", [TIMESTAMP, int(time.time())])
def test_handle_initiate_l1_invalidation(timestamp):
    context, config = make_context()
    context.finalization_times[TARGET_CHAIN_ID] = 1_000

    request = make_request()
    context.requests.add(request.id, request)

    event = InitiateL1InvalidationEvent(
        chain_id=TARGET_CHAIN_ID,
        claim_id=CLAIM_ID,
    )

    # Without a claim, we simply drop the event
    assert process_event(event, context) == (True, None)

    claim = make_claim_challenged(request, claimer=config.account.address)
    context.claims.add(claim.id, claim)

    # Must only be called if claim is invalidated
    with pytest.raises(AssertionError, match="Claim not invalidated"):
        process_event(event, context)

    # Check that task is added to resolution pool
    context.web3_l1.eth.gas_price = Wei(1)  # type: ignore
    claim.start_challenge(make_tx_hash(), timestamp)

    if timestamp == TIMESTAMP:
        assert process_event(event, context) == (True, None)
        assert context.task_pool.submit.called  # type: ignore  # pylint:disable=no-member
    else:
        assert process_event(event, context) == (False, None)
        assert not context.task_pool.submit.called  # type: ignore  # pylint:disable=no-member



import brownie
import pytest
from brownie import chain, web3
from brownie.convert import to_bytes
from eth_utils import to_hex

from beamer.tests.agent.utils import make_address
from beamer.tests.constants import FILL_ID, RM_C_FIELD_TERMINATION
from beamer.tests.util import (
    alloc_accounts,
    create_fill_hash,
    create_request_hash,
    earnings,
    make_request,
)
from beamer.typing import ClaimId, FillId, Termination


def test_request_invalid_target_chain(request_manager, token):
    (requester,) = alloc_accounts(1)
    with brownie.reverts("Target rollup not supported"):
        make_request(
            request_manager,
            target_chain_id=999,
            token=token,
            requester=requester,
            target_address=requester,
            amount=1,
        )

    assert request_manager.requestCounter() == 0
    make_request(
        request_manager,
        target_chain_id=web3.eth.chain_id,
        token=token,
        requester=requester,
        target_address=requester,
        amount=1,
    )
    assert request_manager.requestCounter() == 1


def test_claim(token, request_manager, claim_stake):
    """Test that making a claim creates correct claim and emits event"""
    (requester,) = alloc_accounts(1)
    request_id = make_request(
        request_manager, token=token, requester=requester, target_address=requester, amount=1
    )

    claim_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": requester, "value": claim_stake}
    )
    claim_id = claim_tx.return_value
    expected_termination = (
        request_manager.claimPeriod() + web3.eth.get_block("latest")["timestamp"]
    )

    assert "ClaimMade" in claim_tx.events
    claim_event = claim_tx.events["ClaimMade"]
    assert claim_event["requestId"] == request_id
    assert claim_event["claimId"] == claim_id
    assert claim_event["claimer"] == requester
    assert claim_event["claimerStake"] == claim_stake
    assert claim_event["lastChallenger"] == brownie.ZERO_ADDRESS
    assert claim_event["challengerStakeTotal"] == 0
    assert claim_event["termination"] == expected_termination
    assert claim_event["fillId"] == to_hex(FILL_ID)


def test_claim_with_different_stakes(token, request_manager, claim_stake):
    """Test that only claims with the correct stake can be submitted"""
    (requester,) = alloc_accounts(1)
    request_id = make_request(request_manager, token, requester, requester, 1)

    claim = request_manager.claimRequest(
        request_id, FILL_ID, {"from": requester, "value": claim_stake}
    )
    assert "ClaimMade" in claim.events

    with brownie.reverts("Invalid stake amount"):
        request_manager.claimRequest(
            request_id, FILL_ID, {"from": requester, "value": claim_stake - 1}
        )

    with brownie.reverts("Invalid stake amount"):
        request_manager.claimRequest(
            request_id, FILL_ID, {"from": requester, "value": claim_stake + 1}
        )

    with brownie.reverts("Invalid stake amount"):
        request_manager.claimRequest(request_id, FILL_ID, {"from": requester})


def test_claim_challenge(request_manager, token, claim_stake):
    """Test challenging a claim"""
    requester, challenger = alloc_accounts(2)
    request_id = make_request(request_manager, token, requester, requester, 1)

    claim = request_manager.claimRequest(
        request_id, FILL_ID, {"from": requester, "value": claim_stake}
    )

    with brownie.reverts("Not enough stake provided"):
        request_manager.challengeClaim(
            claim.return_value, {"from": challenger, "value": claim_stake}
        )

    with brownie.reverts("Cannot challenge own claim"):
        request_manager.challengeClaim(
            claim.return_value, {"from": requester, "value": claim_stake + 1}
        )

    with brownie.reverts("Not enough stake provided"):
        request_manager.challengeClaim(claim.return_value, {"from": challenger})

    # Do a proper challenge
    challenge = request_manager.challengeClaim(
        claim.return_value, {"from": challenger, "value": claim_stake + 1}
    )
    assert "ClaimMade" in challenge.events

    with brownie.reverts("Not eligible to outbid"):
        request_manager.challengeClaim(
            claim.return_value, {"from": challenger, "value": claim_stake + 1}
        )


def test_claim_counter_challenge(request_manager, token, claim_stake):
    """Test counter-challenging a challenge"""
    claimer, challenger, requester = alloc_accounts(3)
    request_id = make_request(request_manager, token, requester, requester, 1)

    claim = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim.return_value

    with brownie.reverts("Not enough stake provided"):
        request_manager.challengeClaim(claim_id, {"from": requester, "value": claim_stake})

    # Do a proper challenge
    request_manager.challengeClaim(claim_id, {"from": challenger, "value": claim_stake + 1})

    # Only the claimer is eligible to outbid the challengers
    with brownie.reverts("Not eligible to outbid"):
        request_manager.challengeClaim(claim_id, {"from": requester})

    # The sender of the last challenge must not be able to challenge again
    with brownie.reverts("Not eligible to outbid"):
        request_manager.challengeClaim(claim_id, {"from": challenger})

    # The other party, in this case the claimer, must be able to re-challenge
    with brownie.reverts("Not enough stake provided"):
        request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake})
    outbid = request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake + 1})
    assert "ClaimMade" in outbid.events

    # Check that claimer is leading and cannot challenge own claim
    with brownie.reverts("Cannot challenge own claim"):
        request_manager.challengeClaim(claim_id, {"from": claimer, "value": 1})

    # The challenger must be able to re-challenge, but must increase the stake
    with brownie.reverts("Not enough stake provided"):
        request_manager.challengeClaim(claim_id, {"from": challenger, "value": claim_stake})
    outbid = request_manager.challengeClaim(
        claim_id, {"from": challenger, "value": claim_stake + 1}
    )
    assert "ClaimMade" in outbid.events


def test_claim_two_challengers(request_manager, token, claim_stake):
    """Test that two different challengers can challenge"""
    claimer, first_challenger, second_challenger, requester = alloc_accounts(4)
    request_id = make_request(request_manager, token, requester, requester, 1)

    claim = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim.return_value

    # First challenger challenges
    outbid = request_manager.challengeClaim(
        claim_id, {"from": first_challenger, "value": claim_stake + 1}
    )
    assert "ClaimMade" in outbid.events

    # Claimer outbids again
    outbid = request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake + 1})
    assert "ClaimMade" in outbid.events

    # Check that claimer cannot be second challenger
    with brownie.reverts("Cannot challenge own claim"):
        request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake + 1})

    # Second challenger challenges
    outbid = request_manager.challengeClaim(
        claim_id, {"from": second_challenger, "value": claim_stake + 1}
    )
    assert "ClaimMade" in outbid.events


def test_claim_period_extension(
    request_manager,
    token,
    claim_stake,
    claim_period,
    finalization_time,
    challenge_period_extension,
):
    """Test the extension of the claim/challenge period"""
    claimer, challenger, requester = alloc_accounts(3)
    request_id = make_request(request_manager, token, requester, requester, 1)

    claim = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim.return_value

    def _get_claim_termination(_claim_id: ClaimId) -> Termination:
        return request_manager.claims(_claim_id)[RM_C_FIELD_TERMINATION]

    assert claim.timestamp + claim_period == _get_claim_termination(claim_id)

    challenge = request_manager.challengeClaim(
        claim_id, {"from": challenger, "value": claim_stake + 1}
    )
    challenge_period = finalization_time + challenge_period_extension

    claim_termination = _get_claim_termination(claim_id)
    assert challenge.timestamp + challenge_period == claim_termination

    # Another challenge with big margin to the end of the termination
    # shouldn't increase the termination
    request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake + 1})

    assert claim_termination == _get_claim_termination(claim_id)

    # Another challenge by challenger also shouldn't increase the end of termination
    request_manager.challengeClaim(claim_id, {"from": challenger, "value": claim_stake + 1})
    assert claim_termination == _get_claim_termination(claim_id)

    # Timetravel close to end of challenge period
    chain.mine(timestamp=_get_claim_termination(claim_id) - 10)

    old_claim_termination = claim_termination
    # Claimer challenges close to the end of challenge
    # Should increase the challenge termination
    challenge = request_manager.challengeClaim(
        claim_id, {"from": claimer, "value": claim_stake + 1}
    )

    new_claim_termination = _get_claim_termination(claim_id)
    assert challenge.timestamp + challenge_period_extension == new_claim_termination
    assert new_claim_termination > old_claim_termination

    # Timetravel close to end of challenge period
    chain.mine(timestamp=_get_claim_termination(claim_id) - 10)

    old_claim_termination = new_claim_termination
    rechallenge = request_manager.challengeClaim(
        claim_id, {"from": challenger, "value": claim_stake + 1}
    )
    new_claim_termination = _get_claim_termination(claim_id)
    assert rechallenge.timestamp + challenge_period_extension == new_claim_termination
    assert new_claim_termination > old_claim_termination

    # Timetravel over the end of challenge period
    chain.mine(timestamp=_get_claim_termination(claim_id) + 1)

    with brownie.reverts("Claim expired"):
        request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake + 1})


def test_withdraw_nonexistent_claim(request_manager):
    """Test withdrawing a non-existent claim"""
    with brownie.reverts("claimId not valid"):
        request_manager.withdraw(1234, {"from": alloc_accounts(1)[0]})


def test_claim_nonexistent_request(request_manager):
    """Test claiming a non-existent request"""
    with brownie.reverts("requestId not valid"):
        request_manager.claimRequest(1234, FILL_ID, {"from": alloc_accounts(1)[0]})


def test_withdraw_without_challenge(request_manager, token, claim_stake, claim_period):
    """Test withdraw when a claim was not challenged"""
    requester, claimer = alloc_accounts(2)
    transfer_amount = 23

    claimer_eth_balance = web3.eth.get_balance(claimer.address)

    token.mint(requester, transfer_amount, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount
    assert token.balanceOf(claimer) == 0

    assert web3.eth.get_balance(request_manager.address) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)
    claim_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim_tx.return_value

    assert web3.eth.get_balance(request_manager.address) == claim_stake
    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance - claim_stake

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id, {"from": claimer})

    # Timetravel after claim period
    chain.mine(timedelta=claim_period)

    # Even if the requester calls withdraw, the deposit goes to the claimer
    withdraw_tx = request_manager.withdraw(claim_id, {"from": requester})
    assert "DepositWithdrawn" in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer) == transfer_amount

    assert web3.eth.get_balance(request_manager.address) == 0
    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance

    # Another withdraw must fail
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim_id, {"from": claimer})


def test_withdraw_with_challenge(
    request_manager, token, claim_stake, finalization_time, challenge_period_extension
):
    """Test withdraw when a claim was challenged, and the challenger won.
    In that case, the request funds must not be paid out to the challenger."""

    requester, claimer, challenger = alloc_accounts(3)
    transfer_amount = 23

    claimer_eth_balance = web3.eth.get_balance(claimer.address)
    challenger_eth_balance = web3.eth.get_balance(challenger.address)

    token.mint(requester, transfer_amount, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount
    assert token.balanceOf(claimer) == 0
    assert token.balanceOf(challenger) == 0

    assert web3.eth.get_balance(request_manager.address) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)
    claim_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim_tx.return_value

    assert token.balanceOf(request_manager.address) == transfer_amount

    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance

    request_manager.challengeClaim(claim_id, {"from": challenger, "value": claim_stake + 1})

    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance - claim_stake - 1

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id, {"from": claimer})

    # Timetravel after challenge period
    chain.mine(timedelta=finalization_time + challenge_period_extension)

    assert web3.eth.get_balance(request_manager.address) == 2 * claim_stake + 1

    # The challenger sent the last bid
    # Even if the requester calls withdraw, the challenge stakes go to the challenger
    # However, the request funds stay in the contract
    withdraw_tx = request_manager.withdraw(claim_id, {"from": challenger})
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert "DepositWithdrawn" not in withdraw_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer) == 0
    assert token.balanceOf(challenger) == 0
    assert token.balanceOf(request_manager.address) == transfer_amount

    assert web3.eth.get_balance(request_manager.address) == 0
    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance + claim_stake

    # Another withdraw must fail
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim_id, {"from": claimer})


def test_withdraw_with_two_claims(deployer, request_manager, token, claim_stake, claim_period):
    """Test withdraw when a request was claimed twice"""
    requester, claimer1, claimer2 = alloc_accounts(3)
    transfer_amount = 23

    claimer1_eth_balance = web3.eth.get_balance(claimer1.address)
    claimer2_eth_balance = web3.eth.get_balance(claimer2.address)

    token.mint(requester, transfer_amount, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount
    assert token.balanceOf(claimer1) == 0
    assert token.balanceOf(claimer2) == 0

    assert web3.eth.get_balance(request_manager.address) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)

    claim1_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer1, "value": claim_stake}
    )
    claim1_id = claim1_tx.return_value

    claim2_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer2, "value": claim_stake}
    )
    claim2_id = claim2_tx.return_value

    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim1_id, {"from": claimer1})

    # Timetravel after claim period
    chain.mine(timedelta=claim_period)

    assert web3.eth.get_balance(request_manager.address) == 2 * claim_stake

    # The first claim gets withdrawn first
    # Even if the requester calls withdraw, the deposit goes to the claimer1
    withdraw1_tx = request_manager.withdraw(claim1_id, {"from": requester})
    assert "DepositWithdrawn" in withdraw1_tx.events
    assert "ClaimWithdrawn" in withdraw1_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer1) == transfer_amount
    assert token.balanceOf(claimer2) == 0

    assert web3.eth.get_balance(request_manager.address) == claim_stake
    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake

    # Another withdraw must fail
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim1_id, {"from": claimer1})

    # The other claim must be withdrawable, but the claim stakes go to the
    # contract owner as it is a false claim but no challenger exists.
    with earnings(web3, deployer) as owner_earnings:
        withdraw2_tx = request_manager.withdraw(claim2_id, {"from": requester})
    assert "DepositWithdrawn" not in withdraw2_tx.events
    assert "ClaimWithdrawn" in withdraw2_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer1) == transfer_amount
    assert token.balanceOf(claimer2) == 0

    # Since there was no challenger, but claim2 was a false claim,
    # stakes go to the contract owner.
    assert owner_earnings() == claim_stake
    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake


@pytest.mark.parametrize("second_fill_id", [FILL_ID, FillId(b"wrong_fill_id")])
def test_withdraw_second_claim_same_claimer_different_fill_ids(
    request_manager, token, claim_stake, claim_period, second_fill_id
):
    """
    Test withdraw with two claims by the same address. First one is successful.
    If the second fill id is also equal to the first, this is an identical claim.
    The claimer should also win.
    If the fill id is different the challenger must win,
    even though the claimer was successful with a different claim and fill id.
    """
    requester, claimer, challenger = alloc_accounts(3)
    transfer_amount = 23

    token.mint(requester, transfer_amount, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount
    assert token.balanceOf(claimer) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)

    claim1_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim1_id = claim1_tx.return_value

    claim2_tx = request_manager.claimRequest(
        request_id, second_fill_id, {"from": claimer, "value": claim_stake}
    )
    claim2_id = claim2_tx.return_value

    challenger_eth_balance = web3.eth.get_balance(challenger.address)
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance
    request_manager.challengeClaim(claim2_id, {"from": challenger, "value": claim_stake + 1})
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance - claim_stake - 1

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim1_id, {"from": claimer})
    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim2_id, {"from": claimer})

    # Timetravel after claim period
    chain.mine(timedelta=claim_period)

    # Withdraw must fail because it was challenged
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim2_id, {"from": claimer})

    current_claimer_eth_balance = web3.eth.get_balance(claimer.address)

    withdraw_tx = request_manager.withdraw(claim1_id, {"from": claimer})
    assert "DepositWithdrawn" in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    assert web3.eth.get_balance(claimer.address) == current_claimer_eth_balance + claim_stake
    assert token.balanceOf(claimer) == transfer_amount

    claim_winner = claimer if second_fill_id == FILL_ID else challenger
    claim_loser = challenger if claimer == claim_winner else claimer

    claim_winner_balance = web3.eth.get_balance(claim_winner.address)
    claim_loser_balance = web3.eth.get_balance(claim_loser.address)

    # Even though the challenge period of claim2 isn't over, the claim can be resolved now.
    withdraw_tx = request_manager.withdraw(claim2_id, {"from": claim_winner})
    assert "DepositWithdrawn" not in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    assert web3.eth.get_balance(claim_winner.address) == claim_winner_balance + 2 * claim_stake + 1
    assert web3.eth.get_balance(claim_loser.address) == claim_loser_balance


def test_withdraw_with_two_claims_and_challenge(request_manager, token, claim_stake, claim_period):
    """Test withdraw when a request was claimed twice and challenged"""
    requester, claimer1, claimer2, challenger = alloc_accounts(4)
    transfer_amount = 23

    claimer1_eth_balance = web3.eth.get_balance(claimer1.address)
    claimer2_eth_balance = web3.eth.get_balance(claimer2.address)
    challenger_eth_balance = web3.eth.get_balance(challenger.address)

    token.mint(requester, transfer_amount, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount
    assert token.balanceOf(claimer1) == 0
    assert token.balanceOf(claimer2) == 0

    assert web3.eth.get_balance(request_manager.address) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)

    claim1_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer1, "value": claim_stake}
    )
    claim1_id = claim1_tx.return_value

    claim2_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer2, "value": claim_stake}
    )
    claim2_id = claim2_tx.return_value

    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance

    request_manager.challengeClaim(claim2_id, {"from": challenger, "value": claim_stake + 1})

    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance - claim_stake - 1

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim1_id, {"from": claimer1})

    # Timetravel after claim period
    chain.mine(timedelta=claim_period)

    assert web3.eth.get_balance(request_manager.address) == 3 * claim_stake + 1

    # The first claim gets withdrawn first
    # Even if the requester calls withdraw, the deposit goes to the claimer1
    withdraw1_tx = request_manager.withdraw(claim1_id, {"from": requester})
    assert "DepositWithdrawn" in withdraw1_tx.events
    assert "ClaimWithdrawn" in withdraw1_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer1) == transfer_amount
    assert token.balanceOf(claimer2) == 0

    assert web3.eth.get_balance(request_manager.address) == 2 * claim_stake + 1
    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance - claim_stake - 1

    # Another withdraw must fail
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim1_id, {"from": claimer1})

    # The other claim must be withdrawable, but mustn't transfer tokens again
    request_manager.withdraw(claim2_id, {"from": challenger})

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer1) == transfer_amount
    assert token.balanceOf(claimer2) == 0

    assert web3.eth.get_balance(request_manager.address) == 0
    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance + claim_stake


def test_withdraw_with_two_claims_first_unsuccessful_then_successful(
    request_manager, token, claim_stake, claim_period, finalization_time
):
    """Test withdraw when a request was claimed twice. The first claim fails, while the second
    is successful and should be paid out the request funds."""
    requester, claimer1, claimer2, challenger = alloc_accounts(4)
    transfer_amount = 23

    claimer1_eth_balance = web3.eth.get_balance(claimer1.address)
    claimer2_eth_balance = web3.eth.get_balance(claimer2.address)
    challenger_eth_balance = web3.eth.get_balance(challenger.address)

    token.mint(requester, transfer_amount, {"from": requester})
    assert token.balanceOf(requester) == transfer_amount
    assert token.balanceOf(claimer1) == 0
    assert token.balanceOf(claimer2) == 0
    assert token.balanceOf(request_manager.address) == 0

    assert web3.eth.get_balance(request_manager.address) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)

    claim1_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer1, "value": claim_stake}
    )
    claim1_id = claim1_tx.return_value

    claim2_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer2, "value": claim_stake}
    )
    claim2_id = claim2_tx.return_value

    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance

    request_manager.challengeClaim(claim1_id, {"from": challenger, "value": claim_stake + 1})

    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance - claim_stake - 1

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim1_id, {"from": claimer1})

    # Timetravel after claim period
    chain.mine(timedelta=claim_period + finalization_time)

    assert token.balanceOf(request_manager.address) == transfer_amount
    assert web3.eth.get_balance(request_manager.address) == 3 * claim_stake + 1

    # The first claim gets withdrawn first
    # As the challenger wins, no requests funds must be paid out
    withdraw1_tx = request_manager.withdraw(claim1_id, {"from": challenger})
    assert "DepositWithdrawn" not in withdraw1_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer1) == 0
    assert token.balanceOf(claimer2) == 0
    assert token.balanceOf(request_manager.address) == transfer_amount

    assert web3.eth.get_balance(request_manager.address) == claim_stake
    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance - claim_stake
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance + claim_stake

    # Another withdraw must fail
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim1_id, {"from": claimer1})

    # The other claim must be withdrawable, but mustn't transfer the request tokens again
    withdraw2_tx = request_manager.withdraw(claim2_id, {"from": requester})
    assert "ClaimWithdrawn" in withdraw2_tx.events

    assert token.balanceOf(requester) == 0
    assert token.balanceOf(claimer1) == 0
    assert token.balanceOf(claimer2) == transfer_amount
    assert token.balanceOf(request_manager.address) == 0

    assert web3.eth.get_balance(request_manager.address) == 0
    assert web3.eth.get_balance(claimer1.address) == claimer1_eth_balance - claim_stake
    assert web3.eth.get_balance(claimer2.address) == claimer2_eth_balance
    assert web3.eth.get_balance(challenger.address) == challenger_eth_balance + claim_stake


def test_claim_after_withdraw(request_manager, token, claim_stake, claim_period):
    """Test that the same account can not claim a already withdrawn fill again"""
    requester, claimer = alloc_accounts(2)
    request_id = make_request(request_manager, token, requester, requester, 23)

    claim_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim_tx.return_value

    # Timetravel after claim period
    chain.mine(timedelta=claim_period)
    withdraw_tx = request_manager.withdraw(claim_id, {"from": claimer})
    assert "DepositWithdrawn" in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    # Claiming the same request again must fail
    with brownie.reverts("Deposit already withdrawn"):
        request_manager.claimRequest(request_id, FILL_ID, {"from": claimer, "value": claim_stake})


def test_second_claim_after_withdraw(deployer, request_manager, token, claim_stake, claim_period):
    """Test that one can withdraw a claim immediately after the request
    deposit has been withdrawn via another claim."""
    requester, claimer1, claimer2 = alloc_accounts(3)
    request_id = make_request(request_manager, token, requester, requester, 23)

    claimer1_eth_balance = web3.eth.get_balance(claimer1.address)
    claimer2_eth_balance = web3.eth.get_balance(claimer2.address)

    claim1_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer1, "value": claim_stake}
    )
    claim1_id = claim1_tx.return_value

    # Timetravel after claim period / 2.
    chain.mine(timedelta=claim_period / 2)
    claim2_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer2, "value": claim_stake}
    )
    claim2_id = claim2_tx.return_value

    # Another claim from the future depositReceiver
    claim3_tx = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer1, "value": claim_stake}
    )
    claim3_id = claim3_tx.return_value

    # Timetravel after claim period / 2. At this point claim 1 can be
    # withdrawn (its claim period is over), but not claim 2 (its claim period
    # is not over yet).
    chain.mine(timedelta=claim_period / 2)
    withdraw_tx = request_manager.withdraw(claim1_id, {"from": claimer1})
    assert "DepositWithdrawn" in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert claimer1_eth_balance - claim_stake == web3.eth.get_balance(claimer1.address)

    # Withdrawing the second claim must now succeed immediately because the
    # deposit has been withdrawn and we do not need to wait for the claim
    # period. The stakes go to the contract owner.
    with earnings(web3, deployer) as owner_earnings:
        withdraw_tx = request_manager.withdraw(claim2_id, {"from": claimer2})
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert claimer2_eth_balance - claim_stake == web3.eth.get_balance(claimer2.address)
    assert owner_earnings() == claim_stake

    # Withdrawing the third claim must also succeed immediately.
    # Since the claimer is also the depositReceiver stakes go back to the claimer
    withdraw_tx = request_manager.withdraw(claim3_id, {"from": claimer1})
    assert "ClaimWithdrawn" in withdraw_tx.events
    assert claimer1_eth_balance == web3.eth.get_balance(claimer1.address)


@pytest.mark.parametrize("invalidate", [True, False])
@pytest.mark.parametrize("l1_filler", [make_address(), None])
def test_withdraw_without_challenge_with_resolution(
    request_manager, resolution_registry, token, claim_stake, contracts, invalidate, l1_filler
):
    """
    Test withdraw when a claim was not challenged, but L1 resolved
    It tests the combination of L1 resolution

    fill hash (invalid, valid)
            X
    l1 filler (honest claimer, dishonest claimer)

    In the invalid - dishonest claimer case, stakes go to the contract
    owner as there is no challenger
    In the invalid - honest claimer case, honest claimer reverts the
    invalidation in the resolution registry
    """
    requester, claimer = alloc_accounts(2)
    transfer_amount = 23

    if l1_filler is None:
        l1_filler = claimer.address

    token.mint(requester, transfer_amount, {"from": requester})

    # Initial balances
    claimer_eth_balance = web3.eth.get_balance(claimer.address)
    owner_eth_balance = web3.eth.get_balance(request_manager.owner())
    request_manager_balance = web3.eth.get_balance(request_manager.address)

    requester_token_balance = token.balanceOf(requester)
    claimer_token_balance = token.balanceOf(claimer)

    # If no claims exist or are fully withdrawn
    # there should be no ETH on the request manager contract
    assert web3.eth.get_balance(request_manager.address) == 0

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)

    # Claim
    fill_id = to_bytes(b"123")
    claim_tx = request_manager.claimRequest(
        request_id, fill_id, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim_tx.return_value

    assert web3.eth.get_balance(request_manager.address) == request_manager_balance + claim_stake
    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance - claim_stake

    # Start L1 resolution
    request_hash = create_request_hash(
        request_id,
        web3.eth.chain_id,
        web3.eth.chain_id,
        token.address,
        requester.address,
        transfer_amount,
    )

    fill_hash = create_fill_hash(
        request_id,
        web3.eth.chain_id,
        web3.eth.chain_id,
        token.address,
        requester.address,
        transfer_amount,
        fill_id,
    )

    contracts.l1_messenger.setLastSender(contracts.resolver.address)

    if invalidate:
        resolution_registry.invalidateFillHash(
            request_hash, fill_id, chain.id, {"from": contracts.l1_messenger}
        )
    # Assert that invalidation works
    assert resolution_registry.invalidFillHashes(fill_hash) == invalidate

    # Register a L1 resolution
    resolution_registry.resolveRequest(
        request_hash, fill_id, web3.eth.chain_id, l1_filler, {"from": contracts.l1_messenger}
    )

    # Assert that correct filler is resolved, it reverts the false invalidation
    if invalidate and l1_filler == claimer:
        assert not resolution_registry.invalidFillHashes(fill_hash)

    # The claim period is not over, but the resolution must allow withdrawal now
    withdraw_tx = request_manager.withdraw(claim_id, {"from": claimer})

    if claimer == l1_filler:
        assert "DepositWithdrawn" in withdraw_tx.events
        assert token.balanceOf(requester) == requester_token_balance - transfer_amount
        assert token.balanceOf(claimer) == claimer_token_balance + transfer_amount

    else:
        claimer_eth_balance -= claim_stake
        owner_eth_balance += claim_stake

    assert "ClaimWithdrawn" in withdraw_tx.events

    assert web3.eth.get_balance(request_manager.owner()) == owner_eth_balance
    assert web3.eth.get_balance(claimer.address) == claimer_eth_balance
    assert web3.eth.get_balance(request_manager.address) == request_manager_balance

    # Another withdraw must fail
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim_id, {"from": claimer})


def test_withdraw_l1_resolved_muliple_claims(
    contracts, request_manager, resolution_registry, token, claim_stake
):
    requester, first_claimer, second_claimer = alloc_accounts(3)
    transfer_amount = 23
    token.mint(requester, transfer_amount, {"from": requester})

    # Initial balances
    first_claimer_eth_balance = web3.eth.get_balance(first_claimer.address)
    second_claimer_eth_balance = web3.eth.get_balance(second_claimer.address)
    owner_eth_balance = web3.eth.get_balance(request_manager.owner())

    request_id = make_request(request_manager, token, requester, requester, transfer_amount)

    # Creating 4 Claims
    fill_id = FILL_ID

    # Claim 1: valid claim
    claim_tx_1 = request_manager.claimRequest(
        request_id, fill_id, {"from": first_claimer, "value": claim_stake}
    )
    claim_id_1 = claim_tx_1.return_value

    # Claim 2: claimer is not the filler, invalid claim
    claim_tx_2 = request_manager.claimRequest(
        request_id, fill_id, {"from": second_claimer, "value": claim_stake}
    )
    claim_id_2 = claim_tx_2.return_value

    # Claim 3: another valid claim
    claim_tx_3 = request_manager.claimRequest(
        request_id, fill_id, {"from": first_claimer, "value": claim_stake}
    )
    claim_id_3 = claim_tx_3.return_value

    # Claim 4: claimer is the filler but fill id is wrong, invalid claim
    claim_tx_4 = request_manager.claimRequest(
        request_id, b"wrong fill id", {"from": first_claimer, "value": claim_stake}
    )
    claim_id_4 = claim_tx_4.return_value

    contracts.l1_messenger.setLastSender(contracts.resolver.address)

    # Before L1 resolution, all claims are still running and cannot be withdrawn
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id_1, {"from": first_claimer})
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id_2, {"from": second_claimer})
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id_3, {"from": first_claimer})
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id_4, {"from": first_claimer})

    # Start L1 resolution
    request_hash = create_request_hash(
        request_id,
        web3.eth.chain_id,
        web3.eth.chain_id,
        token.address,
        requester.address,
        transfer_amount,
    )

    # Register a L1 resolution
    resolution_registry.resolveRequest(
        request_hash, fill_id, web3.eth.chain_id, first_claimer, {"from": contracts.l1_messenger}
    )

    # The claim period is not over, but the resolution must allow withdrawal now
    # Valid claim will result in payout
    withdraw_tx = request_manager.withdraw(claim_id_1, {"from": first_claimer})
    assert "DepositWithdrawn" in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    # Wrong claimer, since it is not challenged stakes go to the contract owner
    withdraw_tx = request_manager.withdraw(claim_id_2, {"from": second_claimer})
    assert "DepositWithdrawn" not in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    # Another valid claim, deposit is already withdrawn but stakes go back to claimer
    withdraw_tx = request_manager.withdraw(claim_id_3, {"from": first_claimer})
    assert "DepositWithdrawn" not in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    # Wrong fill id, since it is not challenged stakes go to the contract owner
    withdraw_tx = request_manager.withdraw(claim_id_4, {"from": first_claimer})
    assert "DepositWithdrawn" not in withdraw_tx.events
    assert "ClaimWithdrawn" in withdraw_tx.events

    assert web3.eth.get_balance(first_claimer.address) == first_claimer_eth_balance - claim_stake
    assert web3.eth.get_balance(second_claimer.address) == second_claimer_eth_balance - claim_stake
    # Two of the claims were invalid, thus stakes went to the contract owner
    assert web3.eth.get_balance(request_manager.owner()) == owner_eth_balance + 2 * claim_stake


def test_withdraw_two_challengers(
    request_manager, token, claim_stake, finalization_time, challenge_period_extension
):
    claimer, first_challenger, second_challenger, requester = alloc_accounts(4)

    request_id = make_request(request_manager, token, requester, requester, 1)
    claim = request_manager.claimRequest(
        request_id, FILL_ID, {"from": claimer, "value": claim_stake}
    )
    claim_id = claim.return_value
    first_challenger_eth_balance = web3.eth.get_balance(first_challenger.address)
    second_challenger_eth_balance = web3.eth.get_balance(second_challenger.address)

    # First challenger challenges
    request_manager.challengeClaim(claim_id, {"from": first_challenger, "value": claim_stake + 1})
    # Claimer outbids again
    request_manager.challengeClaim(claim_id, {"from": claimer, "value": claim_stake + 10})
    # Second challenger challenges
    request_manager.challengeClaim(
        claim_id, {"from": second_challenger, "value": claim_stake + 11}
    )

    first_challenger_reward = claim_stake + 1
    second_challenger_reward = claim_stake + 9

    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id, {"from": first_challenger})
    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id, {"from": second_challenger})
    # Withdraw must fail when claim period is not over
    with brownie.reverts("Claim period not finished"):
        request_manager.withdraw(claim_id, {"from": claimer})

    # Timetravel after claim period
    chain.mine(timedelta=finalization_time + challenge_period_extension)

    # Take snapshot
    chain.snapshot()

    def _withdraw_by_order(first_withdrawer, second_withdrawer):
        request_manager.withdraw(claim_id, {"from": first_withdrawer})

        # Challenger cannot withdraw twice
        with brownie.reverts("Challenger has nothing to withdraw"):
            request_manager.withdraw(claim_id, {"from": first_withdrawer})
        with brownie.reverts("Challenger has nothing to withdraw"):
            request_manager.withdraw(claim_id, {"from": claimer})

        request_manager.withdraw(claim_id, {"from": second_withdrawer})

        assert (
            web3.eth.get_balance(first_challenger.address)
            == first_challenger_eth_balance + first_challenger_reward
        )
        assert (
            web3.eth.get_balance(second_challenger.address)
            == second_challenger_eth_balance + second_challenger_reward
        )

    _withdraw_by_order(first_challenger, second_challenger)
    # revert to snapshot
    chain.revert()
    _withdraw_by_order(second_challenger, first_challenger)

    # All stakes are withdrawn already
    with brownie.reverts("Claim already withdrawn"):
        request_manager.withdraw(claim_id, {"from": claimer})


def test_withdraw_expired(token, request_manager):
    """Test that a request can be withdrawn once it is expired"""
    validity_period = 60 * 5
    (requester,) = alloc_accounts(1)

    amount = 17
    token.mint(requester, amount)

    request_id = make_request(
        request_manager, token, requester, requester, amount, validity_period=validity_period
    )

    assert token.balanceOf(requester) == 0

    chain.mine(timedelta=validity_period)
    tx = request_manager.withdrawExpiredRequest(request_id, {"from": requester})
    assert "DepositWithdrawn" in tx.events
    assert token.balanceOf(requester) == amount


def test_withdraw_before_expiration(token, request_manager):
    """Test that a request cannot be withdrawn before it is expired"""
    validity_period = 60 * 5
    (requester,) = alloc_accounts(1)

    request_id = make_request(
        request_manager, token, requester, requester, 1, validity_period=validity_period
    )

    chain.mine(timedelta=validity_period / 2)
    with brownie.reverts("Request not expired yet"):
        request_manager.withdrawExpiredRequest(request_id, {"from": requester})


def test_deprecation(deployer, request_manager, token):
    (requester,) = alloc_accounts(1)
    amount = 17
    token.mint(requester, 2 * amount)

    make_request(
        request_manager,
        token,
        requester,
        requester,
        amount,
    )
    with brownie.reverts("Ownable: caller is not the owner"):
        request_manager.deprecateContract({"from": requester.address})

    assert not request_manager.deprecated()
    request_manager.deprecateContract({"from": deployer.address})
    assert request_manager.deprecated()

    with brownie.reverts("Contract already deprecated"):
        request_manager.deprecateContract({"from": deployer.address})

    with brownie.reverts("Contract is deprecated"):
        make_request(
            request_manager,
            token,
            requester,
            requester,
            amount,
        )


def test_transfer_limit_update_only_owner(deployer, request_manager):
    (random_guy,) = alloc_accounts(1)
    original_transfer_limit = request_manager.transferLimit.call()
    new_transfer_limit = original_transfer_limit + 1

    with brownie.reverts("Ownable: caller is not the owner"):
        request_manager.updateTransferLimit(new_transfer_limit, {"from": random_guy.address})

    assert request_manager.transferLimit.call() == original_transfer_limit
    request_manager.updateTransferLimit(new_transfer_limit, {"from": deployer.address})
    assert request_manager.transferLimit.call() == new_transfer_limit
    # Also show that transfer limit can be decreased again
    request_manager.updateTransferLimit(original_transfer_limit, {"from": deployer.address})
    assert request_manager.transferLimit.call() == original_transfer_limit


def test_transfer_limit_requests(deployer, request_manager, token):
    (requester,) = alloc_accounts(1)
    transfer_limit = request_manager.transferLimit.call()

    assert token.balanceOf(requester) == 0
    token.mint(requester, transfer_limit)

    make_request(request_manager, token, requester, requester, transfer_limit)

    assert token.balanceOf(requester) == 0
    token.mint(requester, transfer_limit + 1)

    with brownie.reverts("Amount exceeds transfer limit"):
        make_request(request_manager, token, requester, requester, transfer_limit + 1)

    request_manager.updateTransferLimit(transfer_limit + 1, {"from": deployer.address})

    make_request(request_manager, token, requester, requester, transfer_limit + 1)

    assert token.balanceOf(requester) == 0



# TODO: Extend and make it smarter to get fields of structs in the contracts
from beamer.typing import FillId

FILL_ID = FillId(b"00000000000000000000000000000abc")
FILL_ID_EMPTY = FillId(bytes(bytearray(32)))

# request fields
RM_R_FIELD_LP_FEE = 9
RM_R_FIELD_PROTOCOL_FEE = 10

# claim fiels
RM_C_FIELD_TERMINATION = 6



import os
import time
from concurrent.futures import Executor, Future
from dataclasses import dataclass, field
from typing import Optional

import structlog
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from statemachine.exceptions import TransitionNotAllowed
from web3 import Web3
from web3.constants import ADDRESS_ZERO
from web3.contract import Contract
from web3.types import BlockData, Timestamp

import beamer.metrics
from beamer.config import Config
from beamer.events import (
    ClaimMade,
    ClaimWithdrawn,
    DepositWithdrawn,
    Event,
    FillHashInvalidated,
    FinalizationTimeUpdated,
    HashInvalidated,
    InitiateL1InvalidationEvent,
    InitiateL1ResolutionEvent,
    LatestBlockUpdatedEvent,
    RequestCreated,
    RequestFilled,
    RequestResolved,
)
from beamer.l1_resolution import run_relayer_for_tx
from beamer.models.claim import Claim
from beamer.models.request import Request
from beamer.tracker import Tracker
from beamer.typing import ChainId, ClaimId, FillHash, RequestHash, RequestId
from beamer.util import TokenMatchChecker

log = structlog.get_logger(__name__)


@dataclass
class Context:
    requests: Tracker[RequestId, Request]
    claims: Tracker[ClaimId, Claim]
    request_manager: Contract
    fill_manager: Contract
    match_checker: TokenMatchChecker
    address: ChecksumAddress
    latest_blocks: dict[ChainId, BlockData]
    config: Config
    web3_l1: Web3
    task_pool: Executor
    l1_resolutions: dict[tuple[RequestId, ClaimId], Future]
    finalization_times: dict[ChainId, int] = field(default_factory=dict)


HandlerResult = tuple[bool, Optional[list[Event]]]


def process_event(event: Event, context: Context) -> HandlerResult:
    log.debug("Processing event", _event=event)

    if isinstance(event, LatestBlockUpdatedEvent):
        return _handle_latest_block_updated(event, context)

    elif isinstance(event, RequestCreated):
        return _handle_request_created(event, context)

    elif isinstance(event, RequestFilled):
        return _handle_request_filled(event, context)

    elif isinstance(event, DepositWithdrawn):
        return _handle_deposit_withdrawn(event, context)

    elif isinstance(event, ClaimMade):
        return _handle_claim_made(event, context)

    elif isinstance(event, ClaimWithdrawn):
        return _handle_claim_withdrawn(event, context)

    elif isinstance(event, RequestResolved):
        return _handle_request_resolved(event, context)

    elif isinstance(event, HashInvalidated):
        return _handle_hash_invalidated(event, context)

    elif isinstance(event, FillHashInvalidated):
        return _handle_fill_hash_invalidated(event, context)

    elif isinstance(event, InitiateL1ResolutionEvent):
        return _handle_initiate_l1_resolution(event, context)

    elif isinstance(event, InitiateL1InvalidationEvent):
        return _handle_initiate_l1_invalidation(event, context)

    elif isinstance(event, FinalizationTimeUpdated):
        return _handle_finalization_time_updated(event, context)

    else:
        raise RuntimeError("Unrecognized event type")


def _find_claims_by_fill_hash(context: Context, fill_hash: FillHash) -> list[Claim]:
    """
    This returns a list, as there can be multiple claims with the same fill hash
    """
    matching_claims = []
    for claim in context.claims:
        request = context.requests.get(claim.request_id)
        assert request is not None

        if fill_hash == request.fill_hash_with_fill_id(claim.fill_id):
            matching_claims.append(claim)

    return matching_claims


def _find_request_by_request_hash(
    context: Context, request_hash: RequestHash
) -> Optional[Request]:
    for request in context.requests:
        if request.request_hash == request_hash:
            return request
    return None


def _invalidation_ready_for_l1_relay(claim: Claim) -> bool:
    return (
        not claim.is_invalidated_l1_resolved
        and claim.invalidation_tx is not None
        and claim.invalidation_timestamp is not None
    )


def _proof_ready_for_l1_relay(request: Request) -> bool:
    return (
        not request.is_l1_resolved
        and request.fill_tx is not None
        and request.fill_timestamp is not None
    )


def _timestamp_is_l1_finalized(
    timestamp: Timestamp, context: Context, target_chain_id: ChainId
) -> bool:
    # The entry in `finalization_times` must exist, because it is checked during request creation
    # target_chain_finalization = context.finalization_times[target_chain_id]

    # FIXME: Remove this for the above once the contracts with the event are deployed
    target_chain_finalization = context.finalization_times.get(target_chain_id)
    if target_chain_finalization is None:
        target_chain_finalization = context.request_manager.functions.finalizationTimes(
            target_chain_id
        ).call()

    return int(time.time()) > timestamp + target_chain_finalization


def _handle_latest_block_updated(
    event: LatestBlockUpdatedEvent, context: Context
) -> HandlerResult:
    context.latest_blocks[event.chain_id] = event.block_data
    return True, None


def _handle_request_created(event: RequestCreated, context: Context) -> HandlerResult:
    # If `BEAMER_ALLOW_UNLISTED_PAIRS` is set, do not check token match file
    if os.environ.get("BEAMER_ALLOW_UNLISTED_PAIRS") is not None:
        # Check if the address points to some contract
        if context.fill_manager.web3.eth.get_code(event.target_token_address) == HexBytes("0x"):
            log.info(
                "Request unfillable, invalid token contract",
                request_event=event,
                token_address=event.target_token_address,
            )
            return True, None
    else:
        is_valid_request = context.match_checker.is_valid_pair(
            event.chain_id,
            event.source_token_address,
            event.target_chain_id,
            event.target_token_address,
        )

        if not is_valid_request:
            log.debug("Invalid token pair in request", _event=event)
            return True, None

    request = Request(
        request_id=event.request_id,
        source_chain_id=event.chain_id,
        target_chain_id=event.target_chain_id,
        source_token_address=event.source_token_address,
        target_token_address=event.target_token_address,
        target_address=event.target_address,
        amount=event.amount,
        valid_until=event.valid_until,
    )
    context.requests.add(request.id, request)

    # We only count valid requests from the agent's perspective to reason about
    # the ratio of fills to valid requests
    with beamer.metrics.update() as data:
        data.requests_created.inc()

    return True, None


def _handle_request_filled(event: RequestFilled, context: Context) -> HandlerResult:
    request = context.requests.get(event.request_id)
    if request is None:
        return False, None

    fill_matches_request = (
        request.id == event.request_id
        and request.amount == event.amount
        and request.source_chain_id == event.source_chain_id
        and request.target_token_address == event.target_token_address
    )
    if not fill_matches_request:
        log.warn("Fill not matching request. Ignoring.", request=request, fill=event)
        return True, None

    try:
        fill_block = context.fill_manager.web3.eth.get_block(event.block_number)
        request.fill(
            filler=event.filler,
            fill_tx=event.tx_hash,
            fill_id=event.fill_id,
            fill_timestamp=fill_block.timestamp,  # type: ignore
        )
    except TransitionNotAllowed:
        return False, None

    with beamer.metrics.update() as data:
        data.requests_filled.inc()
        if event.filler == context.address:
            data.requests_filled_by_agent.inc()

    log.info("Request filled", request=request)
    return True, None


def _handle_deposit_withdrawn(event: DepositWithdrawn, context: Context) -> HandlerResult:
    request = context.requests.get(event.request_id)
    if request is None:
        return True, None

    try:
        request.withdraw()
    except TransitionNotAllowed:
        return False, None

    log.info("Deposit withdrawn", request=request)
    return True, None


def _handle_claim_made(event: ClaimMade, context: Context) -> HandlerResult:
    # RequestCreated event must arrive before ClaimMade
    # Additionally, a request should never be dropped before all claims are finalized
    request = context.requests.get(event.request_id)
    if request is None:
        # We might not find the request because we dropped it for some reason
        # (e.g. an invalid token pair).
        return True, None

    # If we receive a `ClaimMade` event, move forward in the state machine
    if event.claimer == context.address:
        request.try_to_claim()

    claim = context.claims.get(event.claim_id)
    if claim is None:
        challenge_back_off_timestamp = int(time.time())
        # if fill event is not fetched yet, wait `_fill_wait_time`
        # to give the target chain time to sync before challenging
        # additionally, if we are already in the challenge game, no need to back off
        if request.filler is None and event.challenger_stake_total == 0:
            challenge_back_off_timestamp += context.config.fill_wait_time
        claim = Claim(event, challenge_back_off_timestamp)
        context.claims.add(claim.id, claim)

        return True, None

    # Don't process challenge events before there was a chance to invalidate the claim
    if claim.is_started:
        return False, None

    events: list[Event] = []
    if _invalidation_ready_for_l1_relay(claim):
        events.append(
            InitiateL1InvalidationEvent(
                chain_id=request.target_chain_id,  # Resolution happens on the target chain
                claim_id=claim.id,
            )
        )

    # this is at least the second ClaimMade event for this claim id
    assert event.last_challenger != ADDRESS_ZERO, "Second ClaimMade event must contain challenger"
    try:
        claim.challenge(event)
    except TransitionNotAllowed:
        return False, events

    if _proof_ready_for_l1_relay(request):
        events.append(
            InitiateL1ResolutionEvent(
                chain_id=request.target_chain_id,  # Resolution happens on the target chain
                request_id=request.id,
                claim_id=claim.id,
            )
        )

    log.info("Request claimed", request=request, claim_id=event.claim_id)
    return True, events


def _handle_claim_withdrawn(event: ClaimWithdrawn, context: Context) -> HandlerResult:
    claim = context.claims.get(event.claim_id)

    # Check if claim exists, it could happen that we ignored the request because of an
    # invalid token pair, and therefore also did not create the claim
    if claim is None:
        return True, None

    claim.withdraw()
    return True, None


def _handle_request_resolved(event: RequestResolved, context: Context) -> HandlerResult:
    request = _find_request_by_request_hash(context, event.request_hash)
    if request is not None:
        request.l1_resolve(event.filler, event.fill_id)

    return True, None


def _handle_hash_invalidated(event: HashInvalidated, context: Context) -> HandlerResult:
    fill_block = context.fill_manager.web3.eth.get_block(event.block_number)
    claims = _find_claims_by_fill_hash(context, event.fill_hash)

    for claim in claims:
        claim.start_challenge(event.tx_hash, fill_block.timestamp)  # type: ignore

    return True, None


def _handle_fill_hash_invalidated(event: FillHashInvalidated, context: Context) -> HandlerResult:
    claims = _find_claims_by_fill_hash(context, event.fill_hash)
    for claim in claims:
        claim.l1_invalidate()

    return True, None


def _l1_resolution_threshold_reached(claim: Claim, context: Context) -> bool:
    l1_gas_cost = 1_000_000  # TODO: Adapt to real price
    l1_gas_price = context.web3_l1.eth.gas_price
    l1_safety_factor = 1.25
    limit = int(l1_gas_cost * l1_gas_price * l1_safety_factor)

    # Agent is claimer
    if claim.claimer == context.address:
        if claim.latest_claim_made.challenger_stake_total > limit:
            return True
    else:
        # Agent is challenger
        reward = claim.get_challenger_stake(context.address)
        last_challenger = claim.latest_claim_made.last_challenger
        if last_challenger == context.address:
            reward -= (
                claim.latest_claim_made.challenger_stake_total
                - claim.latest_claim_made.claimer_stake
            )

        if reward > limit:
            return True
    return False


def _handle_initiate_l1_resolution(
    event: InitiateL1ResolutionEvent, context: Context
) -> HandlerResult:
    request = context.requests.get(event.request_id)
    claim = context.claims.get(event.claim_id)
    if claim is None:
        return True, None

    # A request should never be dropped before all claims are finalized
    assert request is not None, "Request object missing"
    assert request.fill_tx is not None, "Request not yet filled"
    assert request.fill_timestamp is not None, "Request not yet filled"

    # Check that message is finalized on L1
    if not _timestamp_is_l1_finalized(request.fill_timestamp, context, request.target_chain_id):
        return False, None

    # Check if L1 resolution is cheaper than the winning from challenge
    if _l1_resolution_threshold_reached(claim, context):
        future = context.task_pool.submit(
            run_relayer_for_tx,
            context.config.l1_rpc_url,
            context.config.l2b_rpc_url,
            context.config.account.key,
            request.fill_tx,
        )

        def on_future_done(f: Future) -> None:
            try:
                f.result()
                assert request is not None, "Request object missing"
                assert claim is not None, "Claim object missing"

                del context.l1_resolutions[(request.id, claim.id)]
            except Exception as ex:
                log.error("L1 resolution failed", ex=ex)

        future.add_done_callback(on_future_done)
        context.l1_resolutions[(request.id, claim.id)] = future
        request.l1_resolve()

        log.info("Initiated L1 resolution", request=request, claim=claim)

    return True, None


def _handle_initiate_l1_invalidation(
    event: InitiateL1InvalidationEvent, context: Context
) -> HandlerResult:
    claim = context.claims.get(event.claim_id)
    if claim is None:
        return True, None

    # A request should never be dropped before all claims are finalized
    request = context.requests.get(claim.request_id)
    assert request is not None, "Request object missing"
    assert claim.invalidation_tx is not None, "Claim not invalidated"
    assert claim.invalidation_timestamp is not None, "Claim not invalidated"

    # Check that message is finalized on L1
    if not _timestamp_is_l1_finalized(
        claim.invalidation_timestamp, context, request.target_chain_id
    ):
        return False, None

    # Check if L1 resolution is cheaper than the winning from challenge
    if _l1_resolution_threshold_reached(claim, context):
        future = context.task_pool.submit(
            run_relayer_for_tx,
            context.config.l1_rpc_url,
            context.config.l2b_rpc_url,
            context.config.account.key,
            claim.invalidation_tx,
        )

        def on_future_done(f: Future) -> None:
            try:
                f.result()
                assert request is not None, "Request object missing"
                assert claim is not None, "Claim object missing"

                del context.l1_resolutions[(request.id, claim.id)]
            except Exception as ex:
                log.error("L1 invalidation failed", ex=ex)

        future.add_done_callback(on_future_done)
        context.l1_resolutions[(request.id, claim.id)] = future
        claim.l1_invalidate()

        log.info("Initiated L1 invalidation", request=request, claim=claim)

    return True, None


def _handle_finalization_time_updated(
    event: FinalizationTimeUpdated, context: Context
) -> HandlerResult:
    context.finalization_times[event.target_chain_id] = event.finalization_time
    return True, None
